@@ -5,16 +5,19 @@
 
  */
 
-#define VERSION "1.1.0"
+#define VERSION "1.3.0"
 
 void draw_desktop(void);
 void colorwin(WINDOW * win);
 void printnomem();
+void printipcerr();
 void printkeyhelp(char *keytext, char *desc, WINDOW * win);
 void stdkeyhelp(WINDOW * win);
 void tabkeyhelp(WINDOW * win);
+void scrollkeyhelp();
 void stdexitkeyhelp();
 void menukeyhelp();
 void listkeyhelp();
 void indicate(char *message);
+void printlargenum(unsigned long long i, WINDOW *win);
 void standardcolors(int color);
