@@ -1,22 +1,19 @@
+
 /***
 
 ifstats.c	- the interface statistics module
 Written by Gerard Paul Java
 Copyright (c) Gerard Paul Java 1997
 
-This program is free software; you can redistribute it and/or modify
+This software is open source; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+This program is distributed WITHOUT ANY WARRANTY; without even the
+implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
+See the GNU General Public License in the included COPYING file for
+details.
 	
 ***/
 
@@ -25,9 +22,10 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
-#include <time.h>
+#include <sys/time.h>
 #include <sys/types.h>
 #include <sys/socket.h>
+#include <sys/ioctl.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <net/if_arp.h>
@@ -36,17 +34,21 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 #include <netinet/ip.h>
 #include <netinet/tcp.h>
 #include <netinet/udp.h>
-#include "if_ether.h"
+#include <linux/if_ether.h>
 #include "ifstats.h"
+#include "ifaces.h"
 #include "packet.h"
 #include "log.h"
+#include "dirs.h"
 #include "stdwinset.h"
 #include "deskman.h"
-#include "attrs.h"
 #include "ipcsum.h"
+#include "attrs.h"
 #include "error.h"
 #include "serv.h"
 #include "timer.h"
+#include "options.h"
+#include "instances.h"
 
 #define SCROLLUP 0
 #define SCROLLDOWN 1
@@ -55,15 +57,15 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
 void writegstatlog(struct iftab *table, unsigned long nsecs, FILE * logfile);
 void writedstatlog(char *ifname, float activity, float pps,
-			  float peakactivity, float peakpps,
-			  struct iftotals *ts, unsigned long nsecs,
-			  FILE * logfile);
+		   float peakactivity, float peakpps,
+		   struct iftotals *ts, unsigned long nsecs,
+		   FILE * logfile);
 
 char *ltrim(char *buf)
 {
     char *tmp = buf;
 
-    while (*tmp == ' ')
+    while ((*tmp == ' ') || (*tmp == '\t'))
 	tmp++;
 
     strcpy(buf, tmp);
@@ -91,7 +93,7 @@ int ifinlist(struct iflist *list, char *ifname)
 void initiflist(struct iflist **list)
 {
     FILE *fd;
-    char buf[100];
+    char buf[161];
     char ifname[8];
     struct iflist *itmp = NULL;
     struct iflist *tail = NULL;
@@ -105,15 +107,18 @@ void initiflist(struct iflist **list)
 	errbox("Unable to obtain interface list", ANYKEY_MSG, &resp);
 	return;
     }
-    fgets(buf, 80, fd);		/* dummy header lines */
-    fgets(buf, 80, fd);
+    fgets(buf, 160, fd);	/* dummy header lines */
+    fgets(buf, 160, fd);
 
     while (!(feof(fd))) {
 	strcpy(buf, "");
-	fgets(buf, 80, fd);
+	fgets(buf, 160, fd);
 	if (strcmp(buf, "") != 0) {
 	    strcpy(ifname, ltrim(strtok(buf, ":")));
 
+	    if (!(iface_supported(ifname)))
+		continue;
+
 	    if (ifinlist(*list, ifname))	/* ignore entry if already in */
 		continue;	/* interface list */
 
@@ -138,12 +143,12 @@ void initiflist(struct iflist **list)
     fclose(fd);
 }
 
-void positionptr(struct iflist *list, struct iflist **ptmp, char *ifname)
+void positionptr(struct iftab *table, struct iflist **ptmp, char *ifname)
 {
     struct iflist *plast = NULL;
     int ok = 0;
 
-    *ptmp = list;
+    *ptmp = table->head;
 
     while ((*ptmp != NULL) && (!ok)) {
 	ok = (strcmp((*ptmp)->ifname, ifname) == 0);
@@ -164,6 +169,9 @@ void positionptr(struct iflist *list, struct iflist **ptmp, char *ifname)
 	(*ptmp)->prev_entry = plast;
 	(*ptmp)->next_entry = NULL;
 	strcpy((*ptmp)->ifname, ifname);
+	
+	if ((*ptmp)->index <= LINES - 4)
+	   table->lastvisible = *ptmp;
     }
 }
 
@@ -172,15 +180,17 @@ void destroyiflist(struct iflist *list)
     struct iflist *ctmp;
     struct iflist *ptmp;
 
-    ptmp = list;
-    ctmp = ptmp->next_entry;
+    if (list != NULL) {
+	ptmp = list;
+	ctmp = ptmp->next_entry;
 
-    do {
-	free(ptmp);
-	ptmp = ctmp;
-	if (ctmp != NULL)
-	    ctmp = ctmp->next_entry;
-    } while (ptmp != NULL);
+	do {
+	    free(ptmp);
+	    ptmp = ctmp;
+	    if (ctmp != NULL)
+		ctmp = ctmp->next_entry;
+	} while (ptmp != NULL);
+    }
 }
 
 void updaterates(struct iftab *table, time_t starttime, time_t now,
@@ -206,7 +216,7 @@ void printifentry(struct iflist *ptmp, WINDOW * win, unsigned int idx)
 {
     unsigned int target_row;
 
-    if ((ptmp->index < idx) || (ptmp->index > idx + 14))
+    if ((ptmp->index < idx) || (ptmp->index > idx + (LINES - 5)))
 	return;
 
     target_row = ptmp->index - idx;
@@ -216,13 +226,13 @@ void printifentry(struct iflist *ptmp, WINDOW * win, unsigned int idx)
     wprintw(win, "%s", ptmp->ifname);
     wattrset(win, HIGHATTR);
     wmove(win, target_row, 12);
-    wprintw(win, "%8d", ptmp->total);
+    printlargenum(ptmp->total, win);
     wmove(win, target_row, 22);
-    wprintw(win, "%8d", ptmp->iptotal);
+    printlargenum(ptmp->iptotal, win);
     wmove(win, target_row, 32);
-    wprintw(win, "%8d", ptmp->noniptotal);
+    printlargenum(ptmp->noniptotal, win);
     wmove(win, target_row, 42);
-    wprintw(win, "%8d", ptmp->badtotal);
+    wprintw(win, "%8lu", ptmp->badtotal);
 }
 
 void preparescreen(struct iftab *table)
@@ -230,55 +240,58 @@ void preparescreen(struct iftab *table)
     struct iflist *ptmp = table->head;
     unsigned int i = 1;
 
+    unsigned int winht = LINES - 4;
+
     table->firstvisible = table->head;
 
     do {
 	printifentry(ptmp, table->statwin, 1);
 
-	if (i <= 15)
+	if (i <= winht)
 	    table->lastvisible = ptmp;
 
 	ptmp = ptmp->next_entry;
 	i++;
-    } while ((ptmp != NULL) && (i <= 15));
+    } while ((ptmp != NULL) && (i <= winht));
 }
 
 void labelstats(WINDOW * win)
 {
     wmove(win, 0, 1);
     wprintw(win, " Iface ");
-    wmove(win, 0, 15);
+    wmove(win, 0, 16);
     wprintw(win, " Total ");
-    wmove(win, 0, 28);
+    wmove(win, 0, 29);
     wprintw(win, " IP ");
-    wmove(win, 0, 35);
+    wmove(win, 0, 36);
     wprintw(win, " NonIP ");
     wmove(win, 0, 45);
     wprintw(win, " BadIP ");
     wmove(win, 0, 55);
     wprintw(win, " Activity ");
-};
+}
 
 void initiftab(struct iftab *table)
 {
     initiflist(&(table->head));
 
-    table->borderwin = newwin(17, 80, 4, 0);
+    table->borderwin = newwin(LINES - 2, 80, 1, 0);
     table->borderpanel = new_panel(table->borderwin);
 
-    move(24, 1);
+    move(LINES - 1, 1);
+    scrollkeyhelp();
     stdexitkeyhelp();
     wattrset(table->borderwin, BOXATTR);
     box(table->borderwin, ACS_VLINE, ACS_HLINE);
     labelstats(table->borderwin);
-    table->statwin = newwin(15, 78, 5, 1);
+    table->statwin = newwin(LINES - 4, 78, 2, 1);
     table->statpanel = new_panel(table->statwin);
     stdwinset(table->statwin);
     wtimeout(table->statwin, -1);
     wattrset(table->statwin, STDATTR);
     colorwin(table->statwin);
     wattrset(table->statwin, BOXATTR);
-    wmove(table->borderwin, 16, 32);
+    wmove(table->borderwin, LINES - 3, 32);
     wprintw(table->borderwin, " Total, IP, NonIP, and BadIP are packet counts ");
 }
 
@@ -295,7 +308,7 @@ void scrollgstatwin(struct iftab *table, int direction, unsigned int *idx)
 	    table->lastvisible = table->lastvisible->next_entry;
 	    table->firstvisible = table->firstvisible->next_entry;
 	    (*idx)++;
-	    wmove(table->statwin, 14, 0);
+	    wmove(table->statwin, LINES - 5, 0);
 	    scrollok(table->statwin, 0);
 	    wprintw(table->statwin, "%78c", ' ');
 	    scrollok(table->statwin, 1);
@@ -308,9 +321,7 @@ void scrollgstatwin(struct iftab *table, int direction, unsigned int *idx)
 	    table->lastvisible = table->lastvisible->prev_entry;
 	    (*idx)--;
 	    wmove(table->statwin, 0, 0);
-	    scrollok(table->statwin, 0);
 	    wprintw(table->statwin, "%78c", ' ');
-	    scrollok(table->statwin, 1);
 	    printifentry(table->firstvisible, table->statwin, *idx);
 	}
     }
@@ -321,20 +332,26 @@ void pagegstatwin(struct iftab *table, int direction, int *idx)
     int i = 1;
 
     if (direction == SCROLLUP) {
-	while ((i <= 13) && (table->lastvisible->next_entry != NULL)) {
+	while ((i <= LINES - 5) && (table->lastvisible->next_entry != NULL)) {
 	    i++;
 	    scrollgstatwin(table, direction, idx);
 	}
     } else {
-	while ((i <= 13) && (table->firstvisible != table->head)) {
+	while ((i <= LINES - 5) && (table->firstvisible != table->head)) {
 	    i++;
 	    scrollgstatwin(table, direction, idx);
 	}
     }
 }
 
-void ifstats(int logging, time_t logspan)
+
+/*
+ * The general interface statistics function
+ */
+
+void ifstats(const struct OPTIONS *options, int facilitytime)
 {
+    int logging = options->logging;
     struct iftab table;
 
     char buf[8192];
@@ -360,15 +377,27 @@ void ifstats(int logging, time_t logspan)
     int ch;
 
     int endloop = 0;
-    time_t starttime = 0;
-    time_t statbegin = 0;
-    time_t now = 0;
-    time_t startlog = 0;
+    struct timeval tv;
+    unsigned long starttime = 0;
+    unsigned long statbegin = 0;
+    unsigned long now = 0;
+    unsigned long unow = 0;
+    unsigned long startlog = 0;
+    unsigned long updtime = 0;
+    unsigned long updtime_usec = 0;
 
+    if (!facility_active(GSTATIDFILE))
+        mark_facility(GSTATIDFILE, "general interface statistics");
+    else {
+        errbox("General interface stats already active in another process",
+               ANYKEY_MSG, &ch);
+        return;
+    }
+        
     initiftab(&table);
 
     if (logging) {
-	logfile = opentlog();
+	opentlog(&logfile, GSTATLOG);
 
 	if (logfile == NULL)
 	    logging = 0;
@@ -376,7 +405,7 @@ void ifstats(int logging, time_t logspan)
     writelog(logging, logfile, "******** General interface statistics started");
 
     if (table.head != NULL) {
-	open_socket(&fd);
+	open_socket(&fd, "");
 
 	if (fd < 0)
 	    return;
@@ -386,7 +415,8 @@ void ifstats(int logging, time_t logspan)
 	update_panels();
 	doupdate();
 
-	starttime = startlog = statbegin = time((time_t *) NULL);
+	gettimeofday(&tv, NULL);
+	starttime = startlog = statbegin = tv.tv_sec;
 
 	while (!endloop) {
 	    getpacket(fd, buf, &fromaddr, &ch, &br, 5, table.statwin);
@@ -400,9 +430,11 @@ void ifstats(int logging, time_t logspan)
 		    scrollgstatwin(&table, SCROLLUP, &idx);
 		    break;
 		case KEY_PPAGE:
+		case '-':
 		    pagegstatwin(&table, SCROLLDOWN, &idx);
 		    break;
 		case KEY_NPAGE:
+		case ' ':
 		    pagegstatwin(&table, SCROLLUP, &idx);
 		    break;
 		case 'Q':
@@ -417,7 +449,7 @@ void ifstats(int logging, time_t logspan)
 	    }
 	    if (br > 0) {
 		strcpy(ifname, fromaddr.spkt_device);
-		positionptr(table.head, &ptmp, ifname);
+		positionptr(&table, &ptmp, ifname);
 
 		ptmp->total++;
 
@@ -450,18 +482,29 @@ void ifstats(int logging, time_t logspan)
 		}
 		printifentry(ptmp, table.statwin, idx);
 	    }
-	    now = time((time_t *) NULL);
+	    
+	    gettimeofday(&tv, NULL);
+	    now = tv.tv_sec;
+	    unow = tv.tv_usec;
+	    
 	    if ((now - starttime) >= 5) {
 		updaterates(&table, starttime, now, idx);
-		printelapsedtime(statbegin, now, 16, 1, table.borderwin);
-		starttime = now;		
+		printelapsedtime(statbegin, now, LINES - 3, 1, table.borderwin);
+		starttime = now;
 	    }
-	    if (((now - startlog) >= logspan) && (logging)) {
+	    if (((now - startlog) >= options->logspan) && (logging)) {
 		writegstatlog(&table, time((time_t *) NULL) - statbegin, logfile);
 		startlog = now;
 	    }
-	    update_panels();
-	    doupdate();
+	    if (((options->updrate != 0) && (now - updtime >= options->updrate)) ||
+	       ((options->updrate == 0) && (unow - updtime_usec >= DEFAULT_UPDATE_DELAY))) {
+		update_panels();
+		doupdate();
+		updtime = now;
+		updtime_usec = unow;
+	    }
+	    if ((facilitytime != 0) && (((now - statbegin) / 60) >= facilitytime))
+	        endloop = 1;
 	}
 
 	close(fd);
@@ -479,18 +522,57 @@ void ifstats(int logging, time_t logspan)
 	fclose(logfile);
     }
     destroyiflist(table.head);
+    unlink(GSTATIDFILE);
 }
 
-void printdetlabels(WINDOW * win)
+int initialize_brackets(char *ifname, struct iftotals *ttmp)
 {
-    char *sizelabels = "Pkt Size (bytes)    Count";
+    struct ifreq ifr;
+    int fd;
+    int resp;
+    int istat;
+    unsigned int interval;
+    int i;
+    
+    strcpy(ifr.ifr_name, ifname);
+    
+    fd = socket(PF_INET, SOCK_PACKET, htons(ETH_P_ALL));
+    
+    if (fd < 0) {
+        errbox("Unable to open socket for MTU determination", ANYKEY_MSG, &resp);
+        return 1;
+    }
+    
+    istat = ioctl(fd, SIOCGIFMTU, &ifr);
+    
+    close(fd);
+    if (istat < 0) {
+        errbox("Unable to obtain interface MTU", ANYKEY_MSG, &resp);
+	return 1;
+    }
+            
+    ttmp->interval = interval = ifr.ifr_mtu / 16;    /* There are 16 packet size brackets */
+    
+    for (i = 0; i <= 15; i++) {
+        ttmp->brackets[i].floor = interval * i + 1;
+        ttmp->brackets[i].ceil = interval * (i + 1);
+    }
+    
+    ttmp->brackets[15].ceil = ifr.ifr_mtu;
+    return 0;
+}
+
+void printdetlabels(WINDOW * win, struct iftotals *totals)
+{
+    char *sizelabels = "Pkt Size (bytes)        Count";
+    int i;
 
     wattrset(win, BOXATTR);
-    wmove(win, 2, 13);
-    wprintw(win, "Packets     Bytes");
+    wmove(win, 2, 14);
+    wprintw(win, "Packets       Bytes");
     wmove(win, 12, 2);
     wprintw(win, sizelabels);
-    wmove(win, 12, 32);
+    wmove(win, 12, 36);
     wprintw(win, sizelabels);
     wattrset(win, STDATTR);
     wmove(win, 3, 2);
@@ -513,39 +595,25 @@ void printdetlabels(WINDOW * win)
     wmove(win, 8, 40);
     wprintw(win, "IP Checksum Errors:");
 
-    wmove(win, 13, 2);
-    wprintw(win, "1 to 100:");
-    wmove(win, 14, 2);
-    wprintw(win, "101 to 200:");
-    wmove(win, 15, 2);
-    wprintw(win, "201 to 300:");
-    wmove(win, 16, 2);
-    wprintw(win, "301 to 400:");
-    wmove(win, 17, 2);
-    wprintw(win, "401 to 500:");
-    wmove(win, 18, 2);
-    wprintw(win, "501 to 600:");
-    wmove(win, 19, 2);
-    wprintw(win, "601 to 700:");
-    wmove(win, 20, 2);
-    wprintw(win, "701 to 800:");
 
-    wmove(win, 13, 32);
-    wprintw(win, "801 to 900:");
-    wmove(win, 14, 32);
-    wprintw(win, "901 to 1000:");
-    wmove(win, 15, 32);
-    wprintw(win, "1001 to 1100:");
-    wmove(win, 16, 32);
-    wprintw(win, "1101 to 1200:");
-    wmove(win, 17, 32);
-    wprintw(win, "1201 to 1300:");
-    wmove(win, 18, 32);
-    wprintw(win, "1301 to 1400:");
-    wmove(win, 19, 32);
-    wprintw(win, "1401 to 1500:");
-    wmove(win, 20, 32);
-    wprintw(win, "1500+:");
+    for (i = 0; i <= 7; i++) {
+        wattrset(win, STDATTR);
+        wmove(win, i + 13, 2);
+        wprintw(win, "%4u to %4u:", totals->brackets[i].floor, totals->brackets[i].ceil);
+        wmove(win, i+ 13, 23);
+        wattrset(win, HIGHATTR);
+        wprintw(win, "%8lu", 0);
+    }
+
+    for (i = 8; i <= 15; i++) {
+        wattrset(win, STDATTR);
+        wmove(win, i + 5, 36);
+        wprintw(win, "%4u to %4u:", totals->brackets[i].floor, totals->brackets[i].ceil);
+        wmove(win, i + 5, 57);
+        wattrset(win, HIGHATTR);
+        wprintw(win, "%8lu", 0);
+    }
+
     update_panels();
     doupdate();
 }
@@ -557,84 +625,76 @@ void printdetails(struct iftotals *totals, WINDOW * win)
     /* Print totals on the IP protocols */
 
     wmove(win, 3, 12);
-    wprintw(win, "%8d", totals->total);
-    wmove(win, 3, 22);
-    wprintw(win, "%8d", totals->bytestotal);
+    printlargenum(totals->total, win);
+    wmove(win, 3, 24);
+    printlargenum(totals->bytestotal, win);
     wmove(win, 4, 12);
-    wprintw(win, "%8d", totals->iptotal);
-    wmove(win, 4, 22);
-    wprintw(win, "%8d", totals->ipbtotal);
+    printlargenum(totals->iptotal, win);
+    wmove(win, 4, 24);
+    printlargenum(totals->ipbtotal, win);
 
     wmove(win, 5, 12);
-    wprintw(win, "%8d", totals->tcptotal);
-    wmove(win, 5, 22);
-    wprintw(win, "%8d", totals->tcpbtotal);
+    printlargenum(totals->tcptotal, win);
+    wmove(win, 5, 24);
+    printlargenum(totals->tcpbtotal, win);
 
     wmove(win, 6, 12);
-    wprintw(win, "%8d", totals->udptotal);
-    wmove(win, 6, 22);
-    wprintw(win, "%8d", totals->udpbtotal);
+    printlargenum(totals->udptotal, win);
+    wmove(win, 6, 24);
+    printlargenum(totals->udpbtotal, win);
 
     wmove(win, 7, 12);
-    wprintw(win, "%8d", totals->icmptotal);
-    wmove(win, 7, 22);
-    wprintw(win, "%8d", totals->icmpbtotal);
+    printlargenum(totals->icmptotal, win);
+    wmove(win, 7, 24);
+    printlargenum(totals->icmpbtotal, win);
 
     wmove(win, 8, 12);
-    wprintw(win, "%8d", totals->othtotal);
-    wmove(win, 8, 22);
-    wprintw(win, "%8d", totals->othbtotal);
+    printlargenum(totals->othtotal, win);
+    wmove(win, 8, 24);
+    printlargenum(totals->othbtotal, win);
 
     /* Print non-IP totals */
 
     wmove(win, 9, 12);
-    wprintw(win, "%8d", totals->noniptotal);
-    wmove(win, 9, 22);
-    wprintw(win, "%8d", totals->nonipbtotal);
+    printlargenum(totals->noniptotal, win);
+    wmove(win, 9, 24);
+    printlargenum(totals->nonipbtotal, win);
 
     /* Bad packet count */
 
     wmove(win, 8, 60);
-    wprintw(win, "%8d", totals->badtotal);
-
-    /* Packet size totals */
-    wmove(win, 13, 19);
-    wprintw(win, "%8d", totals->max100);
-    wmove(win, 14, 19);
-    wprintw(win, "%8d", totals->max200);
-    wmove(win, 15, 19);
-    wprintw(win, "%8d", totals->max300);
-    wmove(win, 16, 19);
-    wprintw(win, "%8d", totals->max400);
-    wmove(win, 17, 19);
-    wprintw(win, "%8d", totals->max500);
-    wmove(win, 18, 19);
-    wprintw(win, "%8d", totals->max600);
-    wmove(win, 19, 19);
-    wprintw(win, "%8d", totals->max700);
-    wmove(win, 20, 19);
-    wprintw(win, "%8d", totals->max800);
-
-    wmove(win, 13, 49);
-    wprintw(win, "%8d", totals->max900);
-    wmove(win, 14, 49);
-    wprintw(win, "%8d", totals->max1000);
-    wmove(win, 15, 49);
-    wprintw(win, "%8d", totals->max1100);
-    wmove(win, 16, 49);
-    wprintw(win, "%8d", totals->max1200);
-    wmove(win, 17, 49);
-    wprintw(win, "%8d", totals->max1300);
-    wmove(win, 18, 49);
-    wprintw(win, "%8d", totals->max1400);
-    wmove(win, 19, 49);
-    wprintw(win, "%8d", totals->max1500);
-    wmove(win, 20, 49);
-    wprintw(win, "%8d", totals->g1500);
+    wprintw(win, "%8lu", totals->badtotal);
 }
 
-void detstats(char *iface, int logging, time_t logspan)
+void update_size_distrib(unsigned int length, struct iftotals *totals, WINDOW *win)
 {
+    unsigned int i;
+    
+    i = (length - 1) / totals->interval;   /* minus 1 to keep interval
+    					      boundary lengths within the
+    					      proper brackets */
+    
+    if (i > 15)				   /* This is for extras for MTU's not */
+        i = 15;				   /* divisible by 16 */
+        
+    totals->brackets[i].count++;
+    
+    if (i < 8)
+        wmove(win, i + 13, 23);
+    else
+        wmove(win, i + 5, 57);
+        
+    wprintw(win, "%8lu", totals->brackets[i].count);
+}
+
+/*
+ * The detailed interface statistics function
+ */
+ 
+void detstats(char *iface, const struct OPTIONS *options, int facilitytime)
+{
+    int logging = options->logging;
+
     WINDOW *statwin;
     PANEL *statpanel;
 
@@ -648,6 +708,7 @@ void detstats(char *iface, int logging, time_t logspan)
     int fd;
     FILE *logfile = NULL;
     int br;
+    int framelen = 0;
 
     int iphlen;
     int ipck, ipcsum;
@@ -657,17 +718,39 @@ void detstats(char *iface, int logging, time_t logspan)
 
     int ch;
 
+    struct timeval tv;
+    unsigned long updtime = 0;
+    unsigned long updtime_usec = 0;
     int endloop = 0;
-    time_t starttime, now;
-    time_t statbegin, startlog;
-    unsigned int spanbr = 0;
-    unsigned int spanpkt = 0;
+    unsigned long starttime, now;
+    unsigned long statbegin, startlog;
+    unsigned long unow;
+    float spanbr = 0;
+    float spanpkt = 0;
     float activity = 0;
     float peakactivity = 0;
     float pps = 0;
     float peakpps = 0;
 
-    move(24, 1);
+    /*
+     * Mark this facility
+     */
+     
+    if (!facility_active(DSTATIDFILE))
+        mark_facility(DSTATIDFILE, "detailed interface statistics");
+    else {
+        errbox("Detailed interface stats already active in another process",
+               ANYKEY_MSG, &ch);
+        return;
+    }
+    
+    if (!iface_supported(iface)) {
+        err_iface_unsupported();
+        unlink(DSTATIDFILE);
+        return;
+    }
+        
+    move(LINES - 1, 1);
     stdexitkeyhelp();
     statwin = newwin(23, 80, 1, 0);
     statpanel = new_panel(statwin);
@@ -683,27 +766,45 @@ void detstats(char *iface, int logging, time_t logspan)
     doupdate();
 
     bzero(&totals, sizeof(struct iftotals));
-
+    
     if (logging) {
-	logfile = opentlog();
+	opentlog(&logfile, DSTATLOG);
 
 	if (logfile == NULL)
 	    logging = 0;
     }
     writelog(logging, logfile, "******** Detailed interface statistics started");
-    open_socket(&fd);
+    open_socket(&fd, iface);
 
     if (fd < 0)
 	return;
 
-    printdetlabels(statwin);
+    if (initialize_brackets(iface, &totals) != 0) {
+        close(fd);
+        del_panel(statpanel);
+        delwin(statwin);
+        update_panels();
+        doupdate();
+        unlink(DSTATIDFILE);
+        return;
+    }
+   
+    printdetlabels(statwin, &totals);
     printdetails(&totals, statwin);
     update_panels();
     doupdate();
 
     spanbr = 0;
-    starttime = startlog = statbegin = time((time_t *) NULL);
+    
+    gettimeofday(&tv, NULL);
+    starttime = startlog = statbegin = tv.tv_sec;
+    
+    leaveok(statwin, TRUE);
 
+    /*
+     * Data-gathering loop
+     */
+     
     while (!endloop) {
 	getpacket(fd, buf, &fromaddr, &ch, &br, 5, statwin);
 
@@ -723,6 +824,7 @@ void detstats(char *iface, int logging, time_t logspan)
 	    if (strcmp(iface, fromaddr.spkt_device) != 0)
 		continue;
 
+	    framelen = br;
 	    totals.total++;
 	    totals.bytestotal += br;
 
@@ -769,61 +871,31 @@ void detstats(char *iface, int logging, time_t logspan)
 		    totals.othbtotal += iplen;
 		    break;
 		}
-
-		if (iplen <= 100)
-		    totals.max100++;
-		else if (iplen <= 200)
-		    totals.max200++;
-		else if (iplen <= 300)
-		    totals.max300++;
-		else if (iplen <= 400)
-		    totals.max400++;
-		else if (iplen <= 500)
-		    totals.max500++;
-		else if (iplen <= 600)
-		    totals.max600++;
-		else if (iplen <= 700)
-		    totals.max700++;
-		else if (iplen <= 800)
-		    totals.max800++;
-		else if (iplen <= 900)
-		    totals.max900++;
-		else if (iplen <= 1000)
-		    totals.max1000++;
-		else if (iplen <= 1100)
-		    totals.max1100++;
-		else if (iplen <= 1200)
-		    totals.max1200++;
-		else if (iplen <= 1300)
-		    totals.max1300++;
-		else if (iplen <= 1400)
-		    totals.max1400++;
-		else if (iplen <= 1500)
-		    totals.max1500++;
-		else
-		    totals.g1500++;
 	    } else {
 		totals.noniptotal++;
 		totals.nonipbtotal += br;
 	    }
+            update_size_distrib(framelen, &totals, statwin);
 	}
-	printdetails(&totals, statwin);
-
-	now = time((time_t *) NULL);
+	
+	gettimeofday(&tv, NULL);
+	now = tv.tv_sec;
+	unow = tv.tv_usec;
+	
 	if (now - starttime >= 5) {
 	    wattrset(statwin, BOXATTR);
 	    printelapsedtime(statbegin, now, 22, 1, statwin);
-	    activity = (float) ((spanbr * 8 / 1000) / (float) (now - starttime));
+	    activity = (float) (spanbr * 8 / 1000) / (float) (now - starttime);
 	    pps = (float) (spanpkt) / (float) (now - starttime);
 	    spanbr = 0;
 	    spanpkt = 0;
 	    starttime = now;
-	    
+
 	    wattrset(statwin, HIGHATTR);
 	    wmove(statwin, 5, 50);
-	    wprintw(statwin, "%8.2f kbits/sec", activity);
+	    wprintw(statwin, "%8.1f kbits/sec", activity);
 	    wmove(statwin, 6, 50);
-	    wprintw(statwin, "%8.2f packets/sec", pps);
+	    wprintw(statwin, "%8.1f packets/sec", pps);
 
 	    if (activity > peakactivity)
 		peakactivity = activity;
@@ -831,13 +903,22 @@ void detstats(char *iface, int logging, time_t logspan)
 	    if (pps > peakpps)
 		peakpps = pps;
 	}
-	if (((now - startlog) >= logspan) && (logging)) {
+	if (((now - startlog) >= options->logspan) && (logging)) {
 	    writedstatlog(iface, activity, pps, peakactivity, peakpps,
 		    &totals, time((time_t *) NULL) - statbegin, logfile);
 	    startlog = now;
 	}
-	update_panels();
-	doupdate();
+	if (((options->updrate == 0) && (unow - updtime_usec >= DEFAULT_UPDATE_DELAY)) ||
+	   ((options->updrate != 0) && (now - updtime >= options->updrate))) {
+  	    printdetails(&totals, statwin);
+	    update_panels();
+	    doupdate();
+	    updtime_usec = unow;
+	    updtime = now;
+	}
+	
+	if ((facilitytime != 0) && (((now - statbegin) / 60) >= facilitytime))
+	    endloop = 1;
     }
 
     close(fd);
@@ -850,6 +931,7 @@ void detstats(char *iface, int logging, time_t logspan)
     }
     del_panel(statpanel);
     delwin(statwin);
+    unlink(DSTATIDFILE);
     update_panels();
     doupdate();
 }
@@ -932,12 +1014,10 @@ void selectiface(char *ifname, int *aborted)
 		    row--;
 		else {
 		    wscrl(win, -1);
-		    scrollok(win, 0);
 		    wmove(win, 0, 0);
 		    wprintw(win, "%22c", ' ');
 		    wmove(win, 0, 2);
 		    wprintw(win, ptmp->prev_entry->ifname);
-		    scrollok(win, 1);
 		}
 		ptmp = ptmp->prev_entry;
 	    }
