@@ -6,5 +6,5 @@ Copyright (c) Gerard Paul Java 1997
 
 ***/
 
-FILE *opentlog();
+void opentlog(FILE **fd, char *logfilename);
 void writelog(int logging, FILE * fd, char *msg);
