@@ -6,24 +6,25 @@ packet.c - routines to open the raw socket, read socket data and
 Written by Gerard Paul Java
 Copyright (c) Gerard Paul Java 1997, 1998
 
-This program is free software; you can redistribute it and/or modify
+This software is open source; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+This program is distributed WITHOUT ANY WARRANTY; without even the
+implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
+See the GNU General Public License in the included COPYING file for
+details.
 
 ***/
 
+#ifndef _I386_TYPES_H
+#include <asm/types.h>
+#endif
+
 #include <curses.h>
 #include <panel.h>
+#include <string.h>
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
@@ -31,24 +32,39 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 #include <unistd.h>
 #include <net/if_arp.h>
 #include <linux/if_packet.h>
-#include "if_ether.h"
+#include <linux/if_ether.h>
+#include <linux/if_fddi.h>
 #include "deskman.h"
 #include "error.h"
+#include "options.h"
 
-void open_socket(int *fd)
+void open_socket(int *fd, char *ifname)
 {
     int dummy;
+    struct sockaddr si;
 
     *fd = socket(PF_INET, SOCK_PACKET, htons(ETH_P_ALL));
 
-    if (fd < 0) {
+    if (*fd < 0) {
 	errbox("Unable to open raw socket", ANYKEY_MSG, &dummy);
 	return;
     }
+    
+    /*
+     * Bind socket to a specific interface if one specified.
+     */
+     
+    if (strcmp(ifname, "") != 0) {
+        bzero(&si, sizeof(struct sockaddr));
+        si.sa_family = PF_INET;
+        strcpy(si.sa_data, ifname);
+        bind(*fd, &si, sizeof(struct sockaddr));
+    }
 }
 
 void adjustpacket(char *tpacket, unsigned short family, char **packet,
 		  int *readlen)
+
 {
     switch (family) {
     case ARPHRD_ETHER:
@@ -61,9 +77,12 @@ void adjustpacket(char *tpacket, unsigned short family, char **packet,
     case ARPHRD_CSLIP:
     case ARPHRD_SLIP6:
     case ARPHRD_CSLIP6:
-    case ARPHRD_ADAPT:
 	*packet = tpacket;
 	break;
+    case ARPHRD_FDDI:
+        *packet = tpacket + sizeof(struct fddihdr);
+        *readlen -= sizeof(struct fddihdr);
+        break;
     default:
 	*packet = (char *) NULL;	/* return a NULL packet to signal */
 	break;			/* an unrecognized link protocol */
@@ -77,19 +96,22 @@ void getpacket(int fd, char *buf, struct sockaddr_pkt *fromaddr,
     fd_set set;
     struct timeval tv;
     struct timeval *ptv;
+    int ss;
 
     FD_ZERO(&set);
     FD_SET(0, &set);
     FD_SET(fd, &set);
 
     if (timeout > 0) {
-	tv.tv_sec = timeout;
-	tv.tv_usec = 0;
+	tv.tv_sec = 0;
+	tv.tv_usec = DEFAULT_UPDATE_DELAY;
 	ptv = &tv;
     } else
 	ptv = NULL;
 
-    select(fd + 1, &set, 0, 0, ptv);
+    do {
+	ss = select(fd + 1, &set, 0, 0, ptv);
+    } while (ss < 0);
 
     *br = 0;
     *ch = ERR;
@@ -99,7 +121,7 @@ void getpacket(int fd, char *buf, struct sockaddr_pkt *fromaddr,
 	*br = recvfrom(fd, buf, 8192, 0,
 		       (struct sockaddr *) fromaddr, &fromlen);
     }
-    if (FD_ISSET(0, &set)) {
+
+    if (FD_ISSET(0, &set))
 	*ch = wgetch(win);
-    }
 }
