@@ -6,7 +6,7 @@ Written by Gerard Paul Java
 
 ***/
 
-void open_socket(int *fd);
+void open_socket(int *fd, char *ifname);
 void adjustpacket(char *tpacket, unsigned short family, char **packet,
 		  int *readlen);
 void getpacket(int fd, char *buf, struct sockaddr_pkt *fromaddr,
