@@ -4,19 +4,15 @@ serv.c  - TCP/UDP port statistics module
 Written by Gerard Paul Java
 Copyright (c) Gerard Paul Java 1997, 1998
 
-This program is free software; you can redistribute it and/or modify
+This software is open source; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+This program is distributed WITHOUT ANY WARRANTY; without even the
+implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
+See the GNU General Public License in the included COPYING file for
+details.
 
 ***/
 
@@ -25,33 +21,38 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 #include <stdlib.h>
 #include <unistd.h>
 #include <fcntl.h>
-#include <time.h>
+#include <sys/time.h>
+#include <netdb.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/socket.h>
+#include <netinet/in.h>
 #include <netinet/ip.h>
 #include <netinet/udp.h>
 #include <linux/if_packet.h>
-#include "if_ether.h"
+#include <linux/if_ether.h>
 #include "tcphdr.h"
 #include "dirs.h"
 #include "ipcsum.h"
 #include "deskman.h"
 #include "packet.h"
+#include "ipfrag.h"
+#include "ifaces.h"
 #include "stdwinset.h"
 #include "input.h"
 #include "attrs.h"
 #include "serv.h"
+#include "servname.h"
 #include "error.h"
 #include "log.h"
 #include "timer.h"
+#include "options.h"
+#include "instances.h"
+#include "packet.h"
 
 #define SCROLLUP 0
 #define SCROLLDOWN 1
 
-extern void open_socket(int *fd);
-extern void adjustpacket(char *tpacket, unsigned short family, char **packet,
-			 int *readlen);
 extern void writeutslog(struct portlistent *list, unsigned long nsecs, FILE * logfile);
 
 void initportlist(struct portlist *list)
@@ -60,27 +61,27 @@ void initportlist(struct portlist *list)
     list->firstvisible = list->lastvisible = NULL;
     list->count = 0;
 
-    list->borderwin = newwin(14, 80, 6, 0);
+    list->borderwin = newwin(LINES - 2, 80, 1, 0);
     list->borderpanel = new_panel(list->borderwin);
     wattrset(list->borderwin, BOXATTR);
     box(list->borderwin, ACS_VLINE, ACS_HLINE);
 
     wmove(list->borderwin, 0, 1);
     wprintw(list->borderwin, " Proto/Port ");
-    wmove(list->borderwin, 0, 16);
+    wmove(list->borderwin, 0, 22);
     wprintw(list->borderwin, " Pkts ");
-    wmove(list->borderwin, 0, 26);
+    wmove(list->borderwin, 0, 31);
     wprintw(list->borderwin, " Bytes ");
-    wmove(list->borderwin, 0, 34);
+    wmove(list->borderwin, 0, 40);
     wprintw(list->borderwin, " PktsTo ");
-    wmove(list->borderwin, 0, 44);
+    wmove(list->borderwin, 0, 49);
     wprintw(list->borderwin, " BytesTo ");
-    wmove(list->borderwin, 0, 52);
+    wmove(list->borderwin, 0, 58);
     wprintw(list->borderwin, " PktsFrom ");
-    wmove(list->borderwin, 0, 62);
+    wmove(list->borderwin, 0, 67);
     wprintw(list->borderwin, " BytesFrom ");
 
-    list->win = newwin(12, 78, 7, 1);
+    list->win = newwin(LINES - 4, 78, 2, 1);
     list->panel = new_panel(list->win);
 
     stdwinset(list->win);
@@ -92,7 +93,7 @@ void initportlist(struct portlist *list)
 }
 
 struct portlistent *addtoportlist(struct portlist *list, unsigned int protocol,
-				  unsigned int port, int *nomem)
+			    unsigned int port, int *nomem, int servnames)
 {
     struct portlistent *ptemp;
 
@@ -116,7 +117,14 @@ struct portlistent *addtoportlist(struct portlist *list, unsigned int protocol,
     ptemp->next_entry = NULL;
 
     ptemp->protocol = protocol;
-    ptemp->port = port;
+    ptemp->port = port;		/* This is used in checks later. */
+
+    /* 
+     * Obtain appropriate service name
+     */
+
+    servlook(servnames, htons(port), protocol, ptemp->servname, 10);
+
     ptemp->count = ptemp->bcount = 0;
     ptemp->icount = ptemp->ibcount = 0;
     ptemp->ocount = ptemp->obcount = 0;
@@ -124,10 +132,10 @@ struct portlistent *addtoportlist(struct portlist *list, unsigned int protocol,
     list->count++;
     ptemp->idx = list->count;
 
-    if (list->count <= 12)
+    if (list->count <= LINES - 4)
 	list->lastvisible = ptemp;
 
-    wmove(list->borderwin, 13, 1);
+    wmove(list->borderwin, LINES - 3, 1);
     wprintw(list->borderwin, " %u entries ", list->count);
 
     return ptemp;
@@ -138,7 +146,8 @@ int portinlist(struct porttab *table, unsigned int port)
     struct porttab *ptmp = table;
 
     while (ptmp != NULL) {
-	if (ptmp->port == port)
+	if (((ptmp->port_max == 0) && (ptmp->port_min == port)) ||
+	   ((port >= ptmp->port_min) && (port <= ptmp->port_max)))
 	    return 1;
 
 	ptmp = ptmp->next_entry;
@@ -172,24 +181,33 @@ void printportent(struct portlist *list, struct portlistent *entry,
 {
     unsigned int target_row;
 
-    if ((entry->idx < idx) || (entry->idx > idx + 11))
+    if ((entry->idx < idx) || (entry->idx > idx + (LINES - 5)))
 	return;
 
     target_row = entry->idx - idx;
 
-    wattrset(list->win, STDATTR);
     wmove(list->win, target_row, 1);
-    if (entry->protocol == IPPROTO_TCP)
+    if (entry->protocol == IPPROTO_TCP) {
+	wattrset(list->win, STDATTR);
 	wprintw(list->win, "TCP");
-    else if (entry->protocol == IPPROTO_UDP)
+    } else if (entry->protocol == IPPROTO_UDP) {
+	wattrset(list->win, PTRATTR);
 	wprintw(list->win, "UDP");
-
-    wprintw(list->win, "/%u: ", entry->port);
+    }
+    wprintw(list->win, "/%s: ", entry->servname);
     wattrset(list->win, HIGHATTR);
-    wmove(list->win, target_row, 12);
-    wprintw(list->win, "%8lu %10lu %8lu %10lu %8lu %10lu", entry->count,
-	    entry->bcount, entry->icount, entry->ibcount, entry->ocount,
-	    entry->obcount);
+    wmove(list->win, target_row, 17);
+    printlargenum(entry->count, list->win);
+    wmove(list->win, target_row, 27);
+    printlargenum(entry->bcount, list->win);
+    wmove(list->win, target_row, 37);
+    printlargenum(entry->icount, list->win);
+    wmove(list->win, target_row, 47);
+    printlargenum(entry->ibcount, list->win);
+    wmove(list->win, target_row, 57);
+    printlargenum(entry->ocount, list->win);
+    wmove(list->win, target_row, 67);
+    printlargenum(entry->obcount, list->win);
 }
 
 void destroyportlist(struct portlist *list)
@@ -210,7 +228,8 @@ void destroyportlist(struct portlist *list)
 }
 
 void printport(struct portlist *list, unsigned int protocol, unsigned int port,
-	     int br, unsigned int idx, int fromto, struct porttab *ports)
+	     int br, unsigned int idx, int fromto, struct porttab *ports,
+	       int servnames)
 {
     struct portlistent *listent;
     int nomem = 0;
@@ -219,7 +238,7 @@ void printport(struct portlist *list, unsigned int protocol, unsigned int port,
 	listent = inportlist(list, protocol, port);
 
 	if ((listent == NULL) && (!nomem))
-	    listent = addtoportlist(list, protocol, port, &nomem);
+	    listent = addtoportlist(list, protocol, port, &nomem, servnames);
 
 	if (listent == NULL)
 	    return;
@@ -248,9 +267,9 @@ void scrollservwin(struct portlist *table, int direction, int *idx)
 	    table->lastvisible = table->lastvisible->next_entry;
 	    table->firstvisible = table->firstvisible->next_entry;
 	    (*idx)++;
-	    wmove(table->win, 11, 0);
+	    wmove(table->win, LINES - 5, 0);
 	    scrollok(table->win, 0);
-	    wprintw(table->win, "%78c", 32);
+	    wprintw(table->win, "%78c", ' ');
 	    scrollok(table->win, 1);
 	    printportent(table, table->lastvisible, *idx);
 	}
@@ -261,7 +280,7 @@ void scrollservwin(struct portlist *table, int direction, int *idx)
 	    table->firstvisible = table->firstvisible->prev_entry;
 	    (*idx)--;
 	    wmove(table->win, 0, 0);
-	    wprintw(table->win, "%78c", 32);
+	    wprintw(table->win, "%78c", ' ');
 	    printportent(table, table->firstvisible, *idx);
 	}
     }
@@ -272,20 +291,26 @@ void pageservwin(struct portlist *table, int direction, int *idx)
     int i = 1;
 
     if (direction == SCROLLUP) {
-	while ((i <= 9) && (table->lastvisible != table->tail)) {
+	while ((i <= LINES - 7) && (table->lastvisible != table->tail)) {
 	    i++;
 	    scrollservwin(table, direction, idx);
 	}
     } else {
-	while ((i <= 9) && (table->firstvisible != table->head)) {
+	while ((i <= LINES - 7) && (table->firstvisible != table->head)) {
 	    i++;
 	    scrollservwin(table, direction, idx);
 	}
     }
 }
 
-void servmon(char *ifname, struct porttab *ports, int logging, time_t logspan)
+/*
+ * The TCP/UDP service monitor
+ */
+ 
+void servmon(char *ifname, struct porttab *ports,
+             const struct OPTIONS *options, int facilitytime)
 {
+    int logging = options->logging;
     int fd;
     char buf[8192];
     char *ipacket;
@@ -300,8 +325,13 @@ void servmon(char *ifname, struct porttab *ports, int logging, time_t logspan)
 
     unsigned int sport = 0;
     unsigned int dport = 0;
+    unsigned int firstin;    
 
-    time_t starttime, startlog, timeint, now;
+    struct timeval tv;
+    unsigned long starttime, startlog, timeint;
+    unsigned long now, unow;
+    unsigned long updtime = 0;
+    unsigned long updtime_usec = 0;
 
     int csum;
     int ck;
@@ -312,26 +342,51 @@ void servmon(char *ifname, struct porttab *ports, int logging, time_t logspan)
 
     FILE *logfile = NULL;
 
+    /*
+     * Mark this facility
+     */
+     
+    if (!facility_active(TCPUDPIDFILE))
+        mark_facility(TCPUDPIDFILE, "TCP/UDP monitor");
+    else {
+        errbox("TCP/UDP monitor already active in another process", ANYKEY_MSG, &ch);
+	return;
+    }         
+
+    if (!iface_supported(ifname)) {
+        err_iface_unsupported();
+        unlink(TCPUDPIDFILE);
+        return;
+    }
+        
+
     initportlist(&list);
 
-    move(24, 1);
-    printkeyhelp("Up/Down/PgUp/PgDn", "-scroll window  ", stdscr);
+    move(LINES - 1, 1);
+    scrollkeyhelp();
     stdexitkeyhelp();
 
     update_panels();
     doupdate();
 
+    if (options->servnames)
+	setservent(1);
+
     if (logging) {
-	logfile = opentlog();
+	opentlog(&logfile, TCPUDPLOG);
 
 	if (logfile == NULL)
 	    logging = 0;
     }
     writelog(logging, logfile, "******** TCP/UDP service monitor started");
 
-    open_socket(&fd);
+    open_socket(&fd, ifname);
 
-    starttime = startlog = timeint = time((time_t *) NULL);
+    if (fd < 0)
+	return;
+
+    gettimeofday(&tv, NULL);
+    starttime = startlog = timeint = tv.tv_sec;
 
     do {
 	getpacket(fd, buf, &fromaddr, &ch, &br, 5, list.win);
@@ -345,9 +400,11 @@ void servmon(char *ifname, struct porttab *ports, int logging, time_t logspan)
 		scrollservwin(&list, SCROLLUP, &idx);
 		break;
 	    case KEY_PPAGE:
+	    case '-':
 		pageservwin(&list, SCROLLDOWN, &idx);
 		break;
 	    case KEY_NPAGE:
+	    case ' ':
 		pageservwin(&list, SCROLLUP, &idx);
 		break;
 	    case 'q':
@@ -359,7 +416,7 @@ void servmon(char *ifname, struct porttab *ports, int logging, time_t logspan)
 	    }
 	if (br > 0) {
 	    if ((fromaddr.spkt_protocol == ntohs(ETH_P_IP)) &&
-		(strcmp(ifname, fromaddr.spkt_device) == 0)) {
+	       ((strcmp(fromaddr.spkt_device, ifname) == 0))) {
 		adjustpacket(buf, fromaddr.spkt_family, &ipacket,
 			     &br);
 
@@ -374,41 +431,59 @@ void servmon(char *ifname, struct porttab *ports, int logging, time_t logspan)
 		if (ck != csum)
 		    continue;
 
-		tpacket = ipacket + iphlen;
+		/*
+		 * if not first fragment of a datagram, pass it to
+		 * the fragment processor and get the ports from there.
+		 */
 
-		if (((struct iphdr *) ipacket)->protocol == IPPROTO_TCP) {
-		    sport = ntohs(((struct tcphdr *) tpacket)->source);
-		    dport = ntohs(((struct tcphdr *) tpacket)->dest);
-		} else if (((struct iphdr *) ipacket)->protocol == IPPROTO_UDP) {
-		    sport = ntohs(((struct udphdr *) tpacket)->source);
-		    dport = ntohs(((struct udphdr *) tpacket)->dest);
+		if ((ntohs(((struct iphdr *) ipacket)->frag_off) & 0x3fff) != 0) {
+		    br = processfragment((struct iphdr *) ipacket, &sport, &dport, &firstin);
+		    if (!firstin)
+			continue;
+		} else {
+		    tpacket = ipacket + iphlen;
+
+		    if (((struct iphdr *) ipacket)->protocol == IPPROTO_TCP) {
+		        sport = ntohs(((struct tcphdr *) tpacket)->source);
+		        dport = ntohs(((struct tcphdr *) tpacket)->dest);
+		    } else if (((struct iphdr *) ipacket)->protocol == IPPROTO_UDP) {
+		        sport = ntohs(((struct udphdr *) tpacket)->source);
+		        dport = ntohs(((struct udphdr *) tpacket)->dest);
+		    }
 		}
+		
 		if ((((struct iphdr *) ipacket)->protocol == IPPROTO_TCP) ||
-		    (((struct iphdr *) ipacket)->protocol == IPPROTO_UDP)) {
+		 (((struct iphdr *) ipacket)->protocol == IPPROTO_UDP)) {
 		    printport(&list, ((struct iphdr *) ipacket)->protocol,
 		       sport, ntohs(((struct iphdr *) ipacket)->tot_len),
-			      idx, 0, ports);
+			      idx, 0, ports, options->servnames);
 		    printport(&list, ((struct iphdr *) ipacket)->protocol,
 		       dport, ntohs(((struct iphdr *) ipacket)->tot_len),
-			      idx, 1, ports);
+			      idx, 1, ports, options->servnames);
 		}
 	    }
 	}
-		
-	now = time((time_t *) NULL);
-	
+	gettimeofday(&tv, NULL);
+	now = tv.tv_sec;
+	unow = tv.tv_usec;
+
 	if (now - timeint >= 5) {
-	    printelapsedtime(starttime, now, 13, 20, list.borderwin);
+	    printelapsedtime(starttime, now, LINES - 3, 20, list.borderwin);
 	    timeint = now;
 	}
-	    
-	if (((now - startlog) >= logspan) && (logging)) {
+	if (((now - startlog) >= options->logspan) && (logging)) {
 	    writeutslog(list.head, now - starttime, logfile);
 	    startlog = now;
 	}
-	
-	update_panels();
-	doupdate();
+	if (((options->updrate != 0) && (now - updtime >= options->updrate)) ||
+	   ((options->updrate == 0) && (unow - updtime_usec >= DEFAULT_UPDATE_DELAY))) {
+	    update_panels();
+	    doupdate();
+	    updtime = now;
+	    updtime_usec = unow;
+	}
+	if ((facilitytime != 0) && (((now - starttime) / 60) >= facilitytime))
+	    exitloop = 1;
     } while (!exitloop);
 
     if (logging) {
@@ -416,16 +491,21 @@ void servmon(char *ifname, struct porttab *ports, int logging, time_t logspan)
 	writelog(logging, logfile, "******** TCP/UDP service monitor stopped");
 	fclose(logfile);
     }
+    if (options->servnames)
+	endservent();
+
     del_panel(list.panel);
     delwin(list.win);
     del_panel(list.borderpanel);
     delwin(list.borderwin);
+    unlink(TCPUDPIDFILE);
     update_panels();
     doupdate();
     destroyportlist(&list);
+    destroyfraglist();
 }
 
-void portdlg(unsigned int *port, int *aborted, int mode)
+void portdlg(unsigned int *port_min, int *port_max, int *aborted, int mode)
 {
     WINDOW *bw;
     PANEL *bp;
@@ -434,10 +514,10 @@ void portdlg(unsigned int *port, int *aborted, int mode)
 
     struct FIELDLIST list;
 
-    bw = newwin(12, 47, 5, 5);
+    bw = newwin(14, 50, 5, 5);
     bp = new_panel(bw);
 
-    win = newwin(10, 45, 6, 6);
+    win = newwin(12, 48, 6, 6);
     panel = new_panel(win);
 
     wattrset(bw, BOXATTR);
@@ -448,45 +528,45 @@ void portdlg(unsigned int *port, int *aborted, int mode)
     stdwinset(win);
     wtimeout(win, -1);
 
-    if (mode == 0) {
-	wmove(win, 1, 1);
-	wprintw(win, "Port numbers below 1024 are reserved");
-	wmove(win, 2, 1);
-	wprintw(win, "for TCP/IP services, and are normally");
-	wmove(win, 3, 1);
-	wprintw(win, "the only ones monitored by the TCP/UDP");
-	wmove(win, 4, 1);
-	wprintw(win, "statistics module.  If you wish to monitor");
-	wmove(win, 5, 1);
-	wprintw(win, "a higher-numbered port, enter it here.");
-    } else {
-	wmove(win, 1, 1);
-	wprintw(win, "Enter the port number you wish to remove");
-	wmove(win, 2, 1);
-	wprintw(win, "from the list you created.  You cannot");
-	wmove(win, 3, 1);
-	wprintw(win, "remove ports numbered less then 1024");
-	wmove(win, 4, 1);
-	wprintw(win, "though.");
-    }
+    wmove(win, 1, 1);
+    wprintw(win, "Port numbers below 1024 are reserved for");
+    wmove(win, 2, 1);
+    wprintw(win, "TCP/IP services, and are normally the only");
+    wmove(win, 3, 1);
+    wprintw(win, "ones monitored by the TCP/UDP statistics");
+    wmove(win, 4, 1);
+    wprintw(win, "module.  If you wish to monitor a higher-");
+    wmove(win, 5, 1);
+    wprintw(win, "numbered port or range of ports, enter it");
+    wmove(win, 6, 1);
+    wprintw(win, "here.  Fill just the first field for a");
+    wmove(win, 7, 1);
+    wprintw(win, "single port, or both fields for a range.");
 
-    wmove(win, 9, 1);
+    wmove(win, 11, 1);
+    tabkeyhelp(win);
     stdkeyhelp(win);
 
-    initfields(&list, 1, 6, 13, 8);
+    initfields(&list, 1, 20, 15, 8);
+    wattrset(list.fieldwin, STDATTR);
+    wmove(list.fieldwin, 0, 6);
+    wprintw(list.fieldwin, "to");
+    
     addfield(&list, 5, 0, 0, "");
+    addfield(&list, 5, 0, 9, "");
 
     fillfields(&list, aborted);
 
-    if (!(*aborted))
-	*port = atoi(list.list->buf);
+    if (!(*aborted)) {
+	*port_min = atoi(list.list->buf);
+	*port_max = atoi(list.list->nextfield->buf);
+    }
 
     del_panel(bp);
     delwin(bw);
     del_panel(panel);
     delwin(win);
     destroyfields(&list);
-
 }
 
 void saveportlist(struct porttab *table)
@@ -503,29 +583,53 @@ void saveportlist(struct porttab *table)
 	return;
     }
     while (ptmp != NULL) {
-	bw = write(fd, &(ptmp->port), sizeof(unsigned int));
+	bw = write(fd, &(ptmp->port_min), sizeof(unsigned int));
+	bw = write(fd, &(ptmp->port_max), sizeof(unsigned int));
+	
+	if (bw < 0) {
+	    errbox("Unable to write port/range entry", ANYKEY_MSG, &resp);
+	    destroyporttab(table);
+	    close(fd);
+	    return;
+	}
+	
 	ptmp = ptmp->next_entry;
     }
 
     close(fd);
 }
 
+int dup_portentry(struct porttab *table, unsigned int min, unsigned int max)
+{
+    struct porttab *ptmp = table;
+
+    while (ptmp != NULL) {
+	if ((ptmp->port_min == min) && (ptmp->port_max == max))
+	    return 1;
+
+	ptmp = ptmp->next_entry;
+    }
+
+    return 0;
+}
+
 void addmoreports(struct porttab **table)
 {
-    unsigned int port;
+    unsigned int port_min, port_max;
     int aborted;
     int resp;
     struct porttab *ptmp;
 
-    portdlg(&port, &aborted, 0);
+    portdlg(&port_min, &port_max, &aborted, 0);
 
     if (!aborted) {
-	if (portinlist(*table, port))
-	    errbox("Duplicate port entered", ANYKEY_MSG, &resp);
+	if (dup_portentry(*table, port_min, port_max))
+	    errbox("Duplicate port/range entry", ANYKEY_MSG, &resp);
 	else {
 	    ptmp = malloc(sizeof(struct porttab));
 
-	    ptmp->port = port;
+	    ptmp->port_min = port_min;
+	    ptmp->port_max = port_max;
 	    ptmp->prev_entry = NULL;
 	    ptmp->next_entry = *table;
 
@@ -557,11 +661,13 @@ void loadaddports(struct porttab **table)
     do {
 	ptemp = malloc(sizeof(struct porttab));
 
-	br = read(fd, &(ptemp->port), sizeof(unsigned int));
+	br = read(fd, &(ptemp->port_min), sizeof(unsigned int));
+	br = read(fd, &(ptemp->port_max), sizeof(unsigned int));
 
 	if (br < 0) {
 	    errbox("Error reading port list", ANYKEY_MSG, &resp);
 	    close(fd);
+	    destroyporttab(*table);
 	    return;
 	}
 	if (br > 0) {
@@ -583,13 +689,140 @@ void loadaddports(struct porttab **table)
     close(fd);
 }
 
-void delport(struct porttab **table, unsigned int port)
+void displayportentry(struct porttab *ptmp, WINDOW *win)
+{
+    wprintw(win, "%u", ptmp->port_min);
+    if (ptmp->port_max != 0)
+        wprintw(win, " to %u", ptmp->port_max);	    
+}
+
+void displayports(struct porttab **table, WINDOW * win)
 {
     struct porttab *ptmp = *table;
+    short i = 0;
 
-    while ((ptmp->port != port) && (ptmp != NULL))
+    do {
+	wmove(win, i, 2);
+	displayportentry(ptmp, win);
+		    
+	i++;
 	ptmp = ptmp->next_entry;
+    } while ((i < 18) && (ptmp != NULL));
 
+    update_panels();
+    doupdate();
+}
+
+void operate_portselect(struct porttab **table, struct porttab **node,
+			WINDOW * win, int *aborted)
+{
+    int ch = 0;
+    int i = 0;
+
+    int exitloop = 0;
+
+    *node = *table;
+
+    do {
+	wattrset(win, PTRATTR);
+	wmove(win, i, 1);
+	waddch(win, ACS_RARROW);
+	ch = wgetch(win);
+	wmove(win, i, 1);
+	waddch(win, ' ');
+	wattrset(win, STDATTR);
+
+	switch (ch) {
+	case KEY_DOWN:
+	    if ((*node)->next_entry != NULL) {
+		*node = (*node)->next_entry;
+
+		if (i < 17)
+		    i++;
+		else {
+		    wscrl(win, 1);
+		    scrollok(win, 0);
+		    wmove(win, 17, 0);
+		    wprintw(win, "%22c", ' ');
+		    scrollok(win, 1);
+		    wmove(win, 17, 2);
+		    displayportentry((*node), win);
+		}
+	    }
+	    break;
+	case KEY_UP:
+	    if ((*node)->prev_entry != NULL) {
+		*node = (*node)->prev_entry;
+
+		if (i > 0)
+		    i--;
+		else {
+		    wscrl(win, -1);
+		    wmove(win, 0, 0);
+		    wprintw(win, "%22c", ' ');
+		    wmove(win, 0, 2);
+		    displayportentry((*node), win);
+		}
+	    }
+	    break;
+	case 13:
+	    exitloop = 1;
+	    *aborted = 0;
+	    break;
+	case 27:
+	case 24:
+	case 'x':
+	case 'X':
+	case 'q':
+	case 'Q':
+	    exitloop = 1;
+	    *aborted = 1;
+	    break;
+	}
+    } while (!exitloop);
+}
+
+void selectport(struct porttab **table, struct porttab **node, int *aborted)
+{
+    WINDOW *bw;
+    PANEL *bp;
+    WINDOW *win;
+    PANEL *panel;
+    int resp;
+
+    if (*table == NULL) {
+	errbox("No custom ports", ANYKEY_MSG, &resp);
+	return;
+    }
+    bw = newwin(20, 24, 2, 20);
+    bp = new_panel(bw);
+    wattrset(bw, BOXATTR);
+    box(bw, ACS_VLINE, ACS_HLINE);
+    wmove(bw, 0, 1);
+    wprintw(bw, " Select port entry ");
+    
+    win = newwin(18, 22, 3, 21);
+    panel = new_panel(win);
+    stdwinset(win);
+    wattrset(win, STDATTR);
+    colorwin(win);
+
+    listkeyhelp();
+
+    displayports(table, win);
+    operate_portselect(table, node, win, aborted);
+
+    del_panel(panel);
+    delwin(win);
+    del_panel(bp);
+    delwin(bw);
+
+    update_panels();
+    doupdate();
+}
+
+void delport(struct porttab **table, struct porttab *ptmp)
+{
     if (ptmp != NULL) {
 	if (ptmp == *table) {
 	    *table = (*table)->next_entry;
@@ -608,22 +841,17 @@ void delport(struct porttab **table, unsigned int port)
 
 void removeaport(struct porttab **table)
 {
-    unsigned int port;
     unsigned int aborted;
-    int resp;
-
-    portdlg(&port, &aborted, 1);
+    struct porttab *ptmp;
 
+    selectport(table, &ptmp, &aborted);
+    
     if (!aborted) {
-	if (portinlist(*table, port)) {
-	    delport(table, port);
-	    saveportlist(*table);
-	} else
-	    errbox("Unknown port value", ANYKEY_MSG, &resp);
+	delport(table, ptmp);
+	saveportlist(*table);
     }
 }
 
-
 void destroyporttab(struct porttab *table)
 {
     struct porttab *ptemp = table;
