@@ -5,19 +5,15 @@ tcptable.c - table manipulation routines for the IP monitor
 Written by Gerard Paul Java
 Copyright (c) Gerard Paul Java 1997, 1998
 
-This program is free software; you can redistribute it and/or modify
+This software is open source; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+This program is distributed WITHOUT ANY WARRANTY; without even the
+implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
+See the GNU General Public License in the included COPYING file for
+details.
    
 ***/
 
@@ -26,9 +22,11 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 #include "deskman.h"
 #include "attrs.h"
 #include "log.h"
+#include "revname.h"
 #include "rvnamed.h"
 
-int revname(int lookup, struct in_addr *saddr, char *target);
+unsigned int bmaxy = 0;
+unsigned int imaxy = 0;
 
 void setlabels(WINDOW * win, int mode)
 {
@@ -46,12 +44,34 @@ void setlabels(WINDOW * win, int mode)
 	wmove(win, 0, 56);
 	wprintw(win, " Win Size ");
     }
+}
 
+/*
+ * The hash function for the TCP hash table
+ */
+ 
+unsigned int tcp_hash(unsigned long saddr, unsigned int sport,
+                      unsigned long daddr, unsigned int dport,
+                      char *ifname)
+{
+    int i;
+    int ifsum = 0;
+    
+    for (i = 0; i <= strlen(ifname) - 1; i++)
+        ifsum += ifname[i];
+    
+    return ((ifsum + (4 * saddr) + (3 * sport) + 
+            (2 * daddr) + dport) % ENTRIES_IN_HASH_TABLE);
 }
 
 void init_tcp_table(struct tcptable *table)
 {
-    table->borderwin = newwin(15, 80, 1, 0);
+    int i;
+    
+    table->bmaxy = LINES * 0.6;	             /* 60% of total screen */
+    table->imaxy = table->bmaxy - 2;
+
+    table->borderwin = newwin(table->bmaxy, 80, 1, 0);
     table->borderpanel = new_panel(table->borderwin);
 
     wattrset(table->borderwin, BOXATTR);
@@ -72,7 +92,7 @@ void init_tcp_table(struct tcptable *table)
 
     table->head = table->tail = NULL;
     table->firstvisible = table->lastvisible = NULL;
-    table->tcpscreen = newwin(13, 78, 2, 1);
+    table->tcpscreen = newwin(table->imaxy, 78, 2, 1);
     table->tcppanel = new_panel(table->tcpscreen);
     table->closedentries = NULL;
     wattrset(table->tcpscreen, BOXATTR);
@@ -82,21 +102,109 @@ void init_tcp_table(struct tcptable *table)
 
     wtimeout(table->tcpscreen, -1);
     stdwinset(table->tcpscreen);
+    
+    /* Initialize hash table with dummy header cells */
+    
+    for (i = 0; i <= ENTRIES_IN_HASH_TABLE - 1; i++) {
+        table->hash_table[i] = NULL;
+        table->hash_tails[i] = NULL;
+    }
 }
 
-/* add a new entry to the table */
+/*
+ * Add a TCP entry to the hash table.
+ */
+ 
+int add_tcp_hash_entry(struct tcptable *table, struct tcptableent *entry)
+{
+    unsigned int hp;            /* hash position in table */
+    struct tcp_hashentry *ptmp;
+    
+    hp = tcp_hash(entry->saddr.s_addr, entry->sport,
+                  entry->daddr.s_addr, entry->dport, entry->ifname);
+    
+    ptmp = malloc(sizeof(struct tcp_hashentry));
+    bzero(ptmp, sizeof(struct tcp_hashentry));
+    
+    if (ptmp == NULL)      
+        return 1;
+    
+    /*
+     * Add backpointer from screen node to hash node for deletion later
+     * (Actually point to its predecessor coz of the header cell).
+     */
+     
+    entry->hash_node = ptmp;
+
+    /* 
+     * Update hash node and add it to list.
+     */
+     
+    ptmp->tcpnode = entry;
+    ptmp->hp = hp;
+    
+    if (table->hash_table[hp] == NULL) {
+        ptmp->prev_entry = NULL;
+        table->hash_table[hp] = ptmp;
+    }
+    
+    if (table->hash_tails[hp] != NULL) {
+        table->hash_tails[hp]->next_entry = ptmp;
+        ptmp->prev_entry = table->hash_tails[hp];
+    }
+    
+    table->hash_tails[hp] = ptmp;
+    ptmp->next_entry = NULL;
+    
+    return 0;
+}
+
+/*
+ * Delete a hash table node
+ */
+
+void del_tcp_hash_node(struct tcptable *table, struct tcptableent *entry)
+{
+    struct tcp_hashentry *ptmp;
+
+    ptmp = entry->hash_node;        /* ptmp now points to the target */
+    
+    /*
+     * If the targeted node is the last entry, adjust the corresponding tail
+     * pointer to the preceeding node;
+     */
+     
+    if (ptmp->next_entry == NULL)
+        table->hash_tails[ptmp->hp] = ptmp->prev_entry;
+        
+    if (ptmp->prev_entry != NULL)
+        ptmp->prev_entry->next_entry = ptmp->next_entry;
+    else
+        table->hash_table[ptmp->hp] = ptmp->next_entry;    
+    
+    if (ptmp->next_entry != NULL)
+        ptmp->next_entry->prev_entry = ptmp->prev_entry;
+        
+    free(ptmp);
+}
+
+/*
+ * Add a new entry to the TCP screen table
+ */
 
 struct tcptableent *addentry(struct tcptable *table,
 			     unsigned long int saddr,
-			     unsigned long int daddr,
+ 			     unsigned long int daddr,
 			     unsigned int sport, unsigned int dport,
 			     int protocol, char *ifname, int rev_lookup,
-			     int *nomem)
+			     int rvnfd, int servnames, int *nomem)
 {
     struct tcptableent *new_entry;
     struct closedlist *ctemp;
 
-    /* allocate and attach a new node if no closed entries found */
+    /* 
+     * Allocate and attach a new node if no closed entries found
+     */
 
     if (table->closedentries == NULL) {
 	new_entry = malloc(sizeof(struct tcptableent));
@@ -131,15 +239,16 @@ struct tcptableent *addentry(struct tcptable *table,
 	new_entry->next_entry->prev_entry = new_entry;
 	new_entry->next_entry->next_entry = NULL;
 
-	if (new_entry->oth_connection->index <= table->firstvisible->index + 12)
+
+	if (new_entry->oth_connection->index <= table->firstvisible->index + (table->imaxy - 1))
 	    table->lastvisible = new_entry->oth_connection;
-	else if (new_entry->index <= table->firstvisible->index + 12)
+	else if (new_entry->index <= table->firstvisible->index + (table->imaxy - 1))
 	    table->lastvisible = new_entry;
 
 	new_entry->reused = new_entry->oth_connection->reused = 0;
 	table->count++;
 
-	wmove(table->borderwin, 14, 1);
+	wmove(table->borderwin, table->bmaxy - 1, 1);
 	wprintw(table->borderwin, " TCP: %u entries ", table->count);
     } else {
 	/*
@@ -148,6 +257,7 @@ struct tcptableent *addentry(struct tcptable *table,
 	 * variable point to whatever the top is pointing to.  The new_entry's
 	 * oth_connection also points to the reused entry's oth_connection
 	 */
+
 	new_entry = table->closedentries->closedentry;
 	new_entry->oth_connection = table->closedentries->pair;
 
@@ -155,30 +265,68 @@ struct tcptableent *addentry(struct tcptable *table,
 	table->closedentries = table->closedentries->next_entry;
 	free(ctemp);
 	new_entry->reused = new_entry->oth_connection->reused = 1;
+	
+	/*
+	 * Delete the old hash entries for this reallocated node;
+	 */
+        
+        del_tcp_hash_node(table, new_entry);
+        del_tcp_hash_node(table, new_entry->oth_connection);	
     }
 
+    /* 
+     * Fill in address fields with raw IP addresses
+     */
+
     new_entry->saddr.s_addr = new_entry->oth_connection->daddr.s_addr = saddr;
     new_entry->daddr.s_addr = new_entry->oth_connection->saddr.s_addr = daddr;
     new_entry->protocol = protocol;
+
+    /*
+     * Initialize count fields
+     */
+
     new_entry->pcount = new_entry->bcount = 0;
     new_entry->win = new_entry->psize = 0;
     new_entry->oth_connection->pcount = new_entry->oth_connection->bcount = 0;
     new_entry->oth_connection->win = new_entry->oth_connection->psize = 0;
+    
+    /*
+     * Store interface name
+     */
+     
     strcpy(new_entry->ifname, ifname);
     strcpy(new_entry->oth_connection->ifname, ifname);
+    
+    /*
+     * Set raw port numbers
+     */
 
     new_entry->sport = new_entry->oth_connection->dport = ntohs(sport);
     new_entry->dport = new_entry->oth_connection->sport = ntohs(dport);
 
+
     new_entry->stat = new_entry->oth_connection->stat = 0;
 
-    new_entry->s_fstat = revname(rev_lookup, &(new_entry->saddr), new_entry->s_fqdn);
-    new_entry->d_fstat = revname(rev_lookup, &(new_entry->daddr), new_entry->d_fqdn);
+    new_entry->s_fstat = revname(rev_lookup, &(new_entry->saddr),
+				 new_entry->s_fqdn, rvnfd);
+    new_entry->d_fstat = revname(rev_lookup, &(new_entry->daddr),
+				 new_entry->d_fqdn, rvnfd);
+
+    /*
+     * Set port service names (where applicable)
+     */
+
+    servlook(servnames, sport, IPPROTO_TCP, new_entry->s_sname, 5);
+    servlook(servnames, dport, IPPROTO_TCP, new_entry->d_sname, 5);
+
+    strcpy(new_entry->oth_connection->s_sname, new_entry->d_sname);
+    strcpy(new_entry->oth_connection->d_sname, new_entry->s_sname);
 
     strcpy(new_entry->oth_connection->d_fqdn, new_entry->s_fqdn);
     strcpy(new_entry->oth_connection->s_fqdn, new_entry->d_fqdn);
     new_entry->oth_connection->s_fstat = new_entry->d_fstat;
-    new_entry->oth_connection->d_fstat = new_entry->s_fstat; 
+    new_entry->oth_connection->d_fstat = new_entry->s_fstat;
 
     if (new_entry->index < new_entry->oth_connection->index) {
 	new_entry->half_bracket = ACS_ULCORNER;
@@ -192,6 +340,14 @@ struct tcptableent *addentry(struct tcptable *table,
     new_entry->finack = new_entry->oth_connection->finack = 0;
     new_entry->finsent = new_entry->oth_connection->finsent = 0;
     new_entry->partial = new_entry->oth_connection->partial = 0;
+
+    /*
+     * Add entries to hash table
+     */
+    
+    *nomem = add_tcp_hash_entry(table, new_entry);
+    *nomem = add_tcp_hash_entry(table, new_entry->oth_connection);
+    
     return new_entry;
 }
 
@@ -216,11 +372,13 @@ void addtoclosedlist(struct tcptable *table, struct tcptableent *entry,
 
 struct tcptableent *in_table(struct tcptable *table, unsigned long saddr,
 			     unsigned long daddr, unsigned int sport,
-			     unsigned int dport, int *isclosed,
+			     unsigned int dport, char *ifname, int *isclosed,
 			     unsigned int tm, int logging, FILE * logfile,
 			     int *nomem)
 {
-    struct tcptableent *tblptr;
+    struct tcp_hashentry *hashptr;
+    unsigned int hp;
+    
     char msgstring[120];
 
     int hastimeouts = 0;
@@ -231,13 +389,20 @@ struct tcptableent *in_table(struct tcptable *table, unsigned long saddr,
 	*isclosed = 0;
 	return 0;
     }
-    tblptr = table->head;
 
-    do {
-	if ((tblptr->saddr.s_addr == saddr) &&
-	    (tblptr->daddr.s_addr == daddr) &&
-	    (tblptr->sport == sport) &&
-	    (tblptr->dport == dport))
+    /*
+     * Determine hash table index for this set of addresses and ports
+     */
+     
+    hp = tcp_hash(saddr, sport, daddr, dport, ifname);
+    hashptr = table->hash_table[hp];
+     
+    while (hashptr != NULL) {
+	if ((hashptr->tcpnode->saddr.s_addr == saddr) &&
+	    (hashptr->tcpnode->daddr.s_addr == daddr) &&
+	    (hashptr->tcpnode->sport == sport) &&
+	    (hashptr->tcpnode->dport == dport) &&
+	    (strcmp(hashptr->tcpnode->ifname, ifname) == 0))
 	    break;
 
 	now = time(NULL);
@@ -247,30 +412,31 @@ struct tcptableent *in_table(struct tcptable *table, unsigned long saddr,
 	 * find any closed ones.
 	 */
 
-	if ((tm > 0) && ((now - tblptr->lastupdate) / 60 > tm) && (!tblptr->inclosed)) {
-	    addtoclosedlist(table, tblptr, nomem);
+	if ((tm > 0) && ((now - hashptr->tcpnode->lastupdate) / 60 > tm) && (!(hashptr->tcpnode->inclosed))) {
+	    addtoclosedlist(table, hashptr->tcpnode, nomem);
 	    if (!(*nomem))
 		hastimeouts = 1;
 
 	    if (logging) {
-		sprintf(msgstring, "TCP: Connection %s:%u <-> %s:%u idle for %u minutes, marking for timeout",
-			tblptr->s_fqdn, tblptr->sport, tblptr->d_fqdn, tblptr->dport,
+		sprintf(msgstring, "TCP: Connection %s:%s <-> %s:%s idle for %u minutes, marking for timeout",
+			hashptr->tcpnode->s_fqdn, hashptr->tcpnode->s_sname,
+			hashptr->tcpnode->d_fqdn, hashptr->tcpnode->d_sname,
 			tm);
 		writelog(logging, logfile, msgstring);
 	    }
 	}
-	tblptr = tblptr->next_entry;
-    } while (tblptr != NULL);
+	hashptr = hashptr->next_entry;
+    }
 
-    if (tblptr != NULL) {	/* needed to avoid SIGSEGV */
-	if (((tblptr->finsent == 2) || (tblptr->stat & FLAG_RST)) &&
-	    ((tblptr->oth_connection->finsent == 2) ||
-	     (tblptr->oth_connection->stat & FLAG_RST))) {
+    if (hashptr != NULL) {	/* needed to avoid SIGSEGV */
+	if (((hashptr->tcpnode->finsent == 2) || (hashptr->tcpnode->stat & FLAG_RST)) &&
+	    ((hashptr->tcpnode->oth_connection->finsent == 2) ||
+	     (hashptr->tcpnode->oth_connection->stat & FLAG_RST))) {
 	    *isclosed = 1;
 	    return NULL;
 	} else {
 	    *isclosed = 0;
-	    return tblptr;
+	    return hashptr->tcpnode;
 	}
     } else {
 	*isclosed = hastimeouts;
@@ -281,28 +447,42 @@ struct tcptableent *in_table(struct tcptable *table, unsigned long saddr,
 /*
  * Update the TCP status record should an applicable packet arrive.
  */
- 
+
 void updateentry(struct tcptable *table, struct tcptableent *tableentry,
 		 struct tcphdr *transpacket, unsigned long packetlength,
-		 int logging, int revlook, FILE * logfile, int *nomem)
+		 unsigned int bcount, unsigned int fragofs, int logging,
+		 int revlook, int rvnfd, FILE * logfile, int *nomem)
 {
     char msgstring[160];
 
     if (tableentry->s_fstat != RESOLVED) {
-        tableentry->s_fstat = revname(revlook,&(tableentry->saddr),
-                                      tableentry->s_fqdn);
-        strcpy(tableentry->oth_connection->d_fqdn,tableentry->s_fqdn);
-        tableentry->oth_connection->d_fstat = tableentry->s_fstat;
+	tableentry->s_fstat = revname(revlook, &(tableentry->saddr),
+				      tableentry->s_fqdn, rvnfd);
+	strcpy(tableentry->oth_connection->d_fqdn, tableentry->s_fqdn);
+	tableentry->oth_connection->d_fstat = tableentry->s_fstat;
     }
-                                      
     if (tableentry->d_fstat != RESOLVED) {
-        tableentry->d_fstat = revname(revlook,&(tableentry->daddr),
-                                      tableentry->d_fqdn);
-        strcpy(tableentry->oth_connection->s_fqdn,tableentry->d_fqdn);
-        tableentry->oth_connection->s_fstat = tableentry->d_fstat;
+	tableentry->d_fstat = revname(revlook, &(tableentry->daddr),
+				      tableentry->d_fqdn, rvnfd);
+	strcpy(tableentry->oth_connection->s_fqdn, tableentry->d_fqdn);
+	tableentry->oth_connection->s_fstat = tableentry->d_fstat;
     }
-                                      
     tableentry->pcount++;
+    tableentry->bcount += bcount;
+    tableentry->psize = packetlength;
+
+    /*
+     * If this is not the first TCP fragment, skip interpretation of the
+     * TCP header.
+     */
+
+    if ((ntohs(fragofs) & 0x1fff) != 0) {
+	tableentry->lastupdate = tableentry->oth_connection->lastupdate = time(NULL);
+	return;
+    }
+    /*
+     * At this point, we have a TCP header, and we proceed to process it.
+     */
 
     if (tableentry->pcount == 1) {
 	if ((transpacket->syn) || (transpacket->rst))
@@ -310,8 +490,6 @@ void updateentry(struct tcptable *table, struct tcptableent *tableentry,
 	else
 	    tableentry->partial = 1;
     }
-    tableentry->bcount += packetlength;
-    tableentry->psize = packetlength;
     tableentry->win = ntohs(transpacket->window);
 
     tableentry->stat = 0;
@@ -333,27 +511,18 @@ void updateentry(struct tcptable *table, struct tcptableent *tableentry,
 	 */
 
 	if ((tableentry->oth_connection->finsent == 1) &&
-	   (ntohl(transpacket->seq) == tableentry->oth_connection->finack)) {
+	(ntohl(transpacket->seq) == tableentry->oth_connection->finack)) {
 	    tableentry->oth_connection->finsent = 2;
 
 	    if (logging) {
-		sprintf(msgstring, "TCP: %s:%u to %s:%u: FIN acknowleged. %lu packets, %lu bytes sent",
-			tableentry->s_fqdn, tableentry->sport,
-			tableentry->d_fqdn, tableentry->dport,
+		sprintf(msgstring, "TCP: %s:%s to %s:%s: FIN acknowledged. %lu packets, %lu bytes sent",
+			tableentry->s_fqdn, tableentry->s_sname,
+			tableentry->d_fqdn, tableentry->d_sname,
 			tableentry->pcount, tableentry->bcount);
 
 		writelog(logging, logfile, msgstring);
 	    }
 	}
-	/*
-	 * Shall we add this entry to the closed entry list?  If both directions
-	 * have their state indicators set to 2, that's it.
-	 */
-
-	if ((tableentry->finsent == 2) &&
-	    (tableentry->oth_connection->finsent == 2) &&
-	    (!(tableentry->inclosed)))
-	    addtoclosedlist(table, tableentry, nomem);
     }
     /*
      * The closing sequence is similar, but not identical to the TCP close
@@ -392,9 +561,9 @@ void updateentry(struct tcptable *table, struct tcptableent *tableentry,
 	    tableentry->finack = ntohl(transpacket->ack_seq);
 	}
 	if (logging) {
-	    sprintf(msgstring, "TCP: %s:%u to %s:%u: FIN sent pkt size=%u bytes",
-		    tableentry->s_fqdn, tableentry->sport,
-		    tableentry->d_fqdn, tableentry->dport,
+	    sprintf(msgstring, "TCP: %s:%s to %s:%s: FIN sent pkt size=%u bytes",
+		    tableentry->s_fqdn, tableentry->s_sname,
+		    tableentry->d_fqdn, tableentry->d_sname,
 		    tableentry->psize);
 
 	    writelog(logging, logfile, msgstring);
@@ -406,9 +575,9 @@ void updateentry(struct tcptable *table, struct tcptableent *tableentry,
 	    addtoclosedlist(table, tableentry, nomem);
 
 	if (logging) {
-	    sprintf(msgstring, "TCP: %s:%u to %s:%u: Connection reset",
-		    tableentry->s_fqdn, tableentry->sport,
-		    tableentry->d_fqdn, tableentry->dport);
+	    sprintf(msgstring, "TCP: %s:%s to %s:%s: Connection reset",
+		    tableentry->s_fqdn, tableentry->s_sname,
+		    tableentry->d_fqdn, tableentry->d_sname);
 	    writelog(logging, logfile, msgstring);
 	}
     }
@@ -419,6 +588,16 @@ void updateentry(struct tcptable *table, struct tcptableent *tableentry,
 	tableentry->stat |= FLAG_URG;
 
     tableentry->lastupdate = tableentry->oth_connection->lastupdate = time(NULL);
+	/*
+	 * Shall we add this entry to the closed entry list?  If both
+	 * directions have their state indicators set to 2, that's it.
+	 */
+
+    if ((!tableentry->inclosed) &&
+       (((tableentry->finsent == 2) && ((tableentry->oth_connection->finsent == 1) || (tableentry->oth_connection->finsent == 2))) ||
+       ((tableentry->oth_connection->finsent == 2) && ((tableentry->finsent == 1) || (tableentry->finsent == 2)))))
+	addtoclosedlist(table, tableentry, nomem);
+
 }
 
 /*
@@ -430,27 +609,27 @@ void updateentry(struct tcptable *table, struct tcptableent *tableentry,
  *
  * Returns immediately if the entry is not visible in the window.
  */
- 
+
 void clearaddr(struct tcptable *table, struct tcptableent *tableentry,
-                   unsigned int screen_idx)
+	       unsigned int screen_idx)
 {
     unsigned int target_row;
-    
+
     if ((tableentry->index < screen_idx) ||
-	(tableentry->index > screen_idx + 12))
+	(tableentry->index > screen_idx + (table->imaxy - 1)))
 	return;
 
     target_row = (tableentry->index) - screen_idx;
 
     wmove(table->tcpscreen, target_row, 1);
-    wprintw(table->tcpscreen, "%44c",' ');
+    wprintw(table->tcpscreen, "%44c", ' ');
 }
-    
+
 /*
  * Display a TCP connection line.  Returns immediately if the entry is
  * not visible in the window.
  */
- 
+
 void printentry(struct tcptable *table, struct tcptableent *tableentry,
 		unsigned int screen_idx, int mode)
 {
@@ -458,7 +637,7 @@ void printentry(struct tcptable *table, struct tcptableent *tableentry,
     unsigned int target_row;
 
     if ((tableentry->index < screen_idx) ||
-	(tableentry->index > screen_idx + 12))
+	(tableentry->index > screen_idx + (table->imaxy - 1)))
 	return;
 
     target_row = (tableentry->index) - screen_idx;
@@ -468,13 +647,12 @@ void printentry(struct tcptable *table, struct tcptableent *tableentry,
     wattrset(table->tcpscreen, PTRATTR);
     wmove(table->tcpscreen, target_row, 2);
     if (tableentry->reused) {
-        scrollok(table->tcpscreen, 0);
+	scrollok(table->tcpscreen, 0);
 	wprintw(table->tcpscreen, "%76c", 32);
 	scrollok(table->tcpscreen, 1);
 	tableentry->reused = 0;
 	wmove(table->tcpscreen, target_row, 1);
     }
- 
     /* print half of connection indicator bracket */
 
     wmove(table->tcpscreen, target_row, 0);
@@ -485,18 +663,18 @@ void printentry(struct tcptable *table, struct tcptableent *tableentry,
     wattrset(table->tcpscreen, STDATTR);
 
     wmove(table->tcpscreen, target_row, 1);
-    wprintw(table->tcpscreen, "%s:%u", tableentry->s_fqdn, tableentry->sport);
+    wprintw(table->tcpscreen, "%s:%s", tableentry->s_fqdn, tableentry->s_sname);
 
     wmove(table->tcpscreen, target_row, 23);
-    wprintw(table->tcpscreen, "%s:%u", tableentry->d_fqdn, tableentry->dport);
-    
+    wprintw(table->tcpscreen, "%s:%s", tableentry->d_fqdn, tableentry->d_sname);
+
     wattrset(table->tcpscreen, HIGHATTR);
 
     /* 
      * Print packet and byte counts or window size and packet size, depending
      * on the value of mode.
      */
-     
+
     if (mode == 0) {
 	wmove(table->tcpscreen, target_row, 45);
 	if (tableentry->partial)
@@ -574,7 +752,15 @@ void destroytcptable(struct tcptable *table)
     struct tcptableent *c_next_entry;
     struct closedlist *closedtemp;
     struct closedlist *closedtemp_next;
+    struct tcp_hashentry *hashtemp;
+    struct tcp_hashentry *hashtemp_next;
+    
+    unsigned int i;
 
+    /*
+     * Destroy main TCP table
+     */
+     
     if (table->head != NULL) {
 	ctemp = table->head;
 	c_next_entry = table->head->next_entry;
@@ -587,6 +773,11 @@ void destroytcptable(struct tcptable *table)
 		c_next_entry = c_next_entry->next_entry;
 	}
     }
+ 
+    /*
+     * Destroy list of closed entries
+     */
+     
     if (table->closedentries != NULL) {
 	closedtemp = table->closedentries;
 	closedtemp_next = table->closedentries->next_entry;
@@ -598,4 +789,24 @@ void destroytcptable(struct tcptable *table)
 		closedtemp_next = closedtemp_next->next_entry;
 	}
     }
+    
+    /*
+     * Destroy hash table
+     */
+     
+    for (i = 0; i <= ENTRIES_IN_HASH_TABLE - 1; i++) {
+        if (table->hash_table[i] != NULL) {
+            hashtemp = table->hash_table[i];
+            hashtemp_next = table->hash_table[i]->next_entry;
+           
+            while (hashtemp != NULL) {
+                free(hashtemp);
+                hashtemp = hashtemp_next;
+               
+                if (hashtemp_next != NULL)
+                    hashtemp_next = hashtemp_next->next_entry;
+            }
+        }
+    }     
 }
+
