@@ -1,22 +1,19 @@
+
 /***
 
 utfilter.c  - UDP/TCP display filter module
 Written by Gerard Paul Java
 Copyright (c) Gerard Paul Java 1997, 1998
 
-This program is free software; you can redistribute it and/or modify
+This software is open source; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+This program is distributed WITHOUT ANY WARRANTY; without even the
+implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
+See the GNU General Public License in the included COPYING file for
+details.
 
 ***/
 
@@ -31,6 +28,7 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 #include <time.h>
 #include <unistd.h>
 #include <netdb.h>
+#include <ctype.h>
 #include "dirs.h"
 #include "deskman.h"
 #include "attrs.h"
@@ -39,11 +37,12 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 #include "utfilter.h"
 #include "input.h"
 #include "error.h"
+#include "instances.h"
 
 /*
  * Generate a string representation of a number to be used as a name.
  */
- 
+
 void genname(unsigned long n, char *m)
 {
     sprintf(m, "%lu", n);
@@ -54,9 +53,9 @@ void listfileerr(int code)
     int resp;
 
     if (code == 1)
-	errbox("Error opening filter list file", ANYKEY_MSG, &resp);
+	errbox("Error loading TCP/UDP filter list file", ANYKEY_MSG, &resp);
     else
-	errbox("Error writing filter list file", ANYKEY_MSG, &resp);
+	errbox("Error writing TCP/UDP filter list file", ANYKEY_MSG, &resp);
 }
 
 void gethostparams(struct hostparams *hp, int *aborted)
@@ -66,7 +65,7 @@ void gethostparams(struct hostparams *hp, int *aborted)
     WINDOW *dlgwin;
     PANEL *dlgpanel;
 
-    dlgwin = newwin(10, 80, 7, 0);
+    dlgwin = newwin(12, 80, 7, 0);
     dlgpanel = new_panel(dlgwin);
 
     wattrset(dlgwin, BOXATTR);
@@ -86,13 +85,15 @@ void gethostparams(struct hostparams *hp, int *aborted)
     wmove(dlgwin, 6, 2);
     wprintw(dlgwin, "Port:");
     wmove(dlgwin, 8, 2);
+    wprintw(dlgwin, "Include/Exclude (I/E):");
+    wmove(dlgwin, 10, 2);
     tabkeyhelp(dlgwin);
-    wmove(dlgwin, 8, 20);
+    wmove(dlgwin, 10, 20);
     stdkeyhelp(dlgwin);
     update_panels();
     doupdate();
 
-    initfields(&fieldlist, 7, 52, 8, 27);
+    initfields(&fieldlist, 8, 52, 8, 27);
 
     addfield(&fieldlist, 25, 1, 0, "");
     addfield(&fieldlist, 25, 3, 0, "");
@@ -100,6 +101,7 @@ void gethostparams(struct hostparams *hp, int *aborted)
     addfield(&fieldlist, 25, 1, 26, "0.0.0.0");
     addfield(&fieldlist, 25, 3, 26, "0.0.0.0");
     addfield(&fieldlist, 5, 5, 26, "0");
+    addfield(&fieldlist, 1, 7, 0, "I");
 
     dlist = fieldlist.list->nextfield->nextfield->nextfield;
 
@@ -111,6 +113,7 @@ void gethostparams(struct hostparams *hp, int *aborted)
     strcpy(hp->d_fqdn, dlist->buf);
     strcpy(hp->d_mask, dlist->nextfield->buf);
     hp->dport = atoi(dlist->nextfield->nextfield->buf);
+    hp->reverse = toupper(dlist->nextfield->nextfield->nextfield->buf[0]);
 
     destroyfields(&fieldlist);
     del_panel(dlgpanel);
@@ -123,6 +126,36 @@ void gethostparams(struct hostparams *hp, int *aborted)
 
 }
 
+int mark_filter_change(int protocol)
+{
+    int resp;
+    
+    if (protocol == F_TCP) {
+        if (!facility_active(TCPFLTIDFILE))
+            mark_facility(TCPFLTIDFILE, "TCP filter change");
+        else {
+            errbox("TCP filter data file in use; try again later", ANYKEY_MSG, &resp);
+            return 0;
+        }
+    } else {
+        if (!facility_active(UDPFLTIDFILE))
+            mark_facility(UDPFLTIDFILE, "UDP filter change");
+        else {
+            errbox("UDP filter data file in use; try again later", ANYKEY_MSG, &resp);
+            return 0;
+        }
+    }
+    return 1;
+}
+
+void clear_flt_tag(int protocol)
+{
+    if (protocol == F_TCP)
+        unlink(TCPFLTIDFILE);
+    else
+        unlink(UDPFLTIDFILE);
+}
+
 void definefilter(int protocol, int *aborted)
 {
     struct FIELDLIST descfield;
@@ -138,6 +171,13 @@ void definefilter(int protocol, int *aborted)
     WINDOW *dlgwin;
     PANEL *dlgpanel;
 
+    /*
+     * Lock facility
+     */
+     
+    if (!mark_filter_change(protocol))
+        return;
+        
     dlgwin = newwin(7, 42, 9, 9);
     dlgpanel = new_panel(dlgwin);
     wattrset(dlgwin, BOXATTR);
@@ -161,16 +201,19 @@ void definefilter(int protocol, int *aborted)
     update_panels();
     doupdate();
 
-    if (*aborted)
+    if (*aborted) {
+        clear_flt_tag(protocol);
 	return;
+    }
 
     genname(time((time_t *) NULL), fntemp);
     strcat(ffilename, fntemp);
 
     pfd = open(ffilename, O_CREAT | O_WRONLY | O_TRUNC, S_IRUSR | S_IWUSR);
     if (pfd < 0) {
-	errbox("Cannot create filter record file", ANYKEY_MSG, &resp);
+	errbox("Cannot create TCP/UDP filter record file", ANYKEY_MSG, &resp);
 	*aborted = 1;
+        clear_flt_tag(protocol);
 	return;
     }
     do {
@@ -179,8 +222,9 @@ void definefilter(int protocol, int *aborted)
 	    bw = write(pfd, &hp, sizeof(struct hostparams));
 
 	    if (bw < 0) {
-		errbox("Unable to write filter record", ANYKEY_MSG, &resp);
+		errbox("Unable to write TCP/UDP filter record", ANYKEY_MSG, &resp);
 		close(pfd);
+	        clear_flt_tag(protocol);
 		return;
 	    }
 	}
@@ -195,6 +239,7 @@ void definefilter(int protocol, int *aborted)
 
     if (pfd < 0) {
 	listfileerr(1);
+        clear_flt_tag(protocol);
 	return;
     }
     strcpy(ffile.filename, fntemp);
@@ -203,6 +248,8 @@ void definefilter(int protocol, int *aborted)
 	listfileerr(2);
 
     close(pfd);
+
+    clear_flt_tag(protocol);
 }
 
 void init_filter_table(struct filterlist *fl)
@@ -302,49 +349,6 @@ void operate_select(struct ffnode *ffiles, WINDOW * win, struct ffnode **item,
     *item = pptr;
 }
 
-void loadfilterlist(unsigned int protocol, struct ffnode **fltfile)
-{
-    int pfd;
-
-    struct ffnode *ffiles = NULL;
-    struct ffnode *ptemp;
-    struct ffnode *tail = NULL;
-
-    int br;
-
-    if (protocol == F_TCP)
-	pfd = open(TCPFLNAME, O_RDONLY);
-    else
-	pfd = open(UDPFLNAME, O_RDONLY);
-
-    if (pfd < 0) {
-	listfileerr(1);
-	*fltfile = NULL;
-	return;
-    }
-    do {
-	ptemp = malloc(sizeof(struct ffnode));
-	br = read(pfd, &(ptemp->ffe), sizeof(struct filterfileent));
-
-	if (br > 0) {
-	    if (ffiles == NULL) {
-		ffiles = ptemp;
-		ffiles->prev_entry = NULL;
-	    } else {
-		tail->next_entry = ptemp;
-		ptemp->prev_entry = tail;
-	    }
-
-	    ptemp->next_entry = NULL;
-	    tail = ptemp;
-	} else
-	    free(ptemp);
-    } while (br > 0);
-
-    close(pfd);
-    *fltfile = ffiles;
-}
-
 void pickafilter(struct ffnode *ffiles,
 		 struct ffnode **fltfile, int *aborted)
 {
@@ -376,21 +380,6 @@ void pickafilter(struct ffnode *ffiles,
     doupdate();
 }
 
-void destroyfilterlist(struct ffnode *fltlist)
-{
-    struct ffnode *fftemp;
-
-    if (fltlist != NULL) {
-	fftemp = fltlist->next_entry;
-
-	do {
-	    free(fltlist);
-	    fltlist = fftemp;
-	    if (fftemp != NULL)
-		fftemp = fftemp->next_entry;
-	} while (fltlist != NULL);
-    }
-}
 
 void selectfilter(unsigned int protocol,
 		  struct filterfileent *ffe, int *aborted)
@@ -398,10 +387,10 @@ void selectfilter(unsigned int protocol,
     struct ffnode *fltfile;
     struct ffnode *ffiles;
 
-    loadfilterlist(protocol, &ffiles);
-
-    if (ffiles == NULL) {
+    if (loadfilterlist(protocol, &ffiles) == 1) {
+	listfileerr(1);
 	*aborted = 1;
+	destroyfilterlist(ffiles);
 	return;
     }
     pickafilter(ffiles, &fltfile, aborted);
@@ -418,28 +407,27 @@ unsigned long int nametoaddr(char *ascname)
     struct hostent *he;
     char imsg[45];
     int resp;
-    
+
     result = inet_addr(ascname);
     if (result == -1) {
-        bzero(imsg, 45);
-        strcpy(imsg, "Resolving ");    
-        strncat(imsg, ascname, 35);
-        indicate(imsg);
-        he = gethostbyname(ascname);
-        if (he != NULL)
-            bcopy((he->h_addr_list)[0], &result, he->h_length);
-        else {
-            bzero(imsg,45);
-            strcpy(imsg, "Unable to resolve ");
-            strncat(imsg, ascname, 22);
-            errbox(imsg, ANYKEY_MSG, &resp);
-            result = -1;
-        }
+	bzero(imsg, 45);
+	strcpy(imsg, "Resolving ");
+	strncat(imsg, ascname, 35);
+	indicate(imsg);
+	he = gethostbyname(ascname);
+	if (he != NULL)
+	    bcopy((he->h_addr_list)[0], &result, he->h_length);
+	else {
+	    bzero(imsg, 45);
+	    strcpy(imsg, "Unable to resolve ");
+	    strncat(imsg, ascname, 22);
+	    errbox(imsg, ANYKEY_MSG, &resp);
+	    result = -1;
+	}
     }
-    
     return result;
 }
-    
+
 /* loads the filter from the filter file */
 
 void loadfilter(char *filename, struct filterlist *fl)
@@ -455,7 +443,7 @@ void loadfilter(char *filename, struct filterlist *fl)
     pfd = open(filename, O_RDONLY);
 
     if (pfd < 0) {
-	errbox("Error opening filter file", ANYKEY_MSG, &resp);
+	errbox("Error opening TCP/UDP filter file", ANYKEY_MSG, &resp);
 	fl->head = NULL;
 	return;
     }
@@ -465,14 +453,13 @@ void loadfilter(char *filename, struct filterlist *fl)
 
 	if (br > 0) {
 	    fe->index = idx;
-	    fe->saddr = nametoaddr(fe->hp.s_fqdn);	    
+	    fe->saddr = nametoaddr(fe->hp.s_fqdn);
 	    fe->daddr = nametoaddr(fe->hp.d_fqdn);
-	    
+
 	    if ((fe->saddr == -1) || (fe->daddr == -1)) {
-	        free(fe);
-	        continue;
+		free(fe);
+		continue;
 	    }
-	    
 	    fe->smask = inet_addr(fe->hp.s_mask);
 	    fe->dmask = inet_addr(fe->hp.d_mask);
 
@@ -494,6 +481,53 @@ void loadfilter(char *filename, struct filterlist *fl)
 	close(pfd);
 }
 
+void marktcpfilter(char *filename)
+{
+    int fd;
+    int resp;
+
+    fd = open(TCPFILTERSAVE, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
+
+    if (write(fd, filename, 40) < 0) {
+	errbox("Unable to save TCP/UDP filter name", ANYKEY_MSG, &resp);
+
+	if (fd > 0)
+	    close(fd);
+
+	return;
+    }
+    close(fd);
+}
+
+void loadsavedtcpfilter(struct filterlist *fl, int *filtered)
+{
+    int fd;
+    char filename[40];
+    int resp;
+
+    fd = open(TCPFILTERSAVE, O_RDONLY);
+
+    if (fd < 0)
+	return;
+
+    bzero(filename, 40);
+    if (read(fd, filename, 40) < 0) {
+	errbox("Unable to retrieve saved TCP/UDP filter", ANYKEY_MSG, &resp);
+	if (fd > 0)
+	    close(fd);
+
+	*filtered = 0;
+	return;
+    }
+    close(fd);
+
+    loadfilter(filename, fl);
+
+    if (fl->head == NULL)
+	*filtered = 0;
+    else
+	*filtered = 1;
+}
 
 /* the display filter */
 
@@ -519,9 +553,12 @@ int utfilter(struct filterlist *fl,
 	if (((csaddr == fsaddr) && ((fe->hp.sport == sport) || (fe->hp.sport == 0)) &&
 	     (cdaddr == fdaddr) && ((fe->hp.dport == dport) || (fe->hp.dport == 0))) ||
 	    ((crsaddr == fdaddr) && ((fe->hp.dport == sport) || (fe->hp.dport == 0)) &&
-	     (crdaddr == fsaddr) && ((fe->hp.sport == dport) || (fe->hp.sport == 0))))
-	    return 1;
+	     (crdaddr == fsaddr) && ((fe->hp.sport == dport) || (fe->hp.sport == 0)))) {
+	    if (toupper(fe->hp.reverse) == 'E')		/* Exclusive */
+		return 0;
 
+	    return 1;
+	}
 	fe = fe->next_entry;
     }
 
@@ -561,15 +598,21 @@ void delfilter(unsigned int protocol, int *aborted)
     int fd;
     int bw;
 
-    loadfilterlist(protocol, &fltlist);
+    if (!mark_filter_change(protocol))
+        return;
 
-    if (fltlist == NULL)
+    if (loadfilterlist(protocol, &fltlist) == 1) {
+	listfileerr(1);
+	destroyfilterlist(fltlist);
+	clear_flt_tag(protocol);
 	return;
-
+    }
     pickafilter(fltlist, &fltfile, aborted);
 
-    if (*aborted)
+    if (*aborted) {
+	clear_flt_tag(protocol);
 	return;
+    }
 
     strcat(fntemp, fltfile->ffe.filename);
     unlink(fntemp);
@@ -593,7 +636,8 @@ void delfilter(unsigned int protocol, int *aborted)
 	fd = open(UDPFLNAME, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
 
     if (fd < 0) {
-	listfileerr(1);
+	listfileerr(2);
+	clear_flt_tag(protocol);
 	return;
     }
     fltfile = fltlist;
@@ -602,6 +646,7 @@ void delfilter(unsigned int protocol, int *aborted)
 
 	if (bw < 0) {
 	    listfileerr(2);
+	    clear_flt_tag(protocol);
 	    return;
 	}
 	ffntemp = fltfile;
@@ -610,6 +655,8 @@ void delfilter(unsigned int protocol, int *aborted)
     }
 
     close(fd);
+    
+    clear_flt_tag(protocol);
 }
 
 void makefiltermenu(struct MENU *menu)
@@ -714,6 +761,7 @@ void tcpfilterselect(struct filterlist *fl,
 		strcat(filename, ffe.filename);
 		loadfilter(filename, fl);
 		*filtered = 1;
+		marktcpfilter(filename);
 	    }
 	    break;
 	case 3:
@@ -721,6 +769,7 @@ void tcpfilterselect(struct filterlist *fl,
 		destroyfilter(fl);
 		*filtered = 0;
 	    }
+	    unlink(TCPFILTERSAVE);
 	    break;
 	case 4:
 	    delfilter(F_TCP, faborted);
